@@ -1,5 +1,5 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
 
     This file is part of Threading Building Blocks. Threading Building Blocks is free software;
     you can redistribute it and/or modify it under the terms of the GNU General Public License
@@ -24,11 +24,7 @@
 #include "tbb/blocked_range2d.h"
 #include "tbb/task_scheduler_init.h"
 #include "tbb/tick_count.h"
-#if TBB_IMPLEMENT_CPP0X
 #include "tbb/compat/thread"
-#else
-#include <thread>
-#endif
 
 #include <math.h>
 #include <stdio.h>
