@@ -25,10 +25,32 @@
 #error Do not #include this internal file directly; use public TBB headers instead.
 #endif
 
+// included in namespace tbb::flow::interface8 (in flow_graph.h)
+
 namespace internal {
 
+    typedef tbb::internal::uint64_t tag_value;
+
+    using tbb::internal::strip;
+
     namespace graph_policy_namespace {
-        enum graph_buffer_policy { rejecting, reserving, queueing, tag_matching };
+
+        struct rejecting { };
+        struct reserving { };
+        struct queueing  { };
+
+        // K == type of field used for key-matching.  Each tag-matching port will be provided
+        // functor that, given an object accepted by the port, will return the
+        /// field of type K being used for matching.
+        template<typename K, typename KHash=tbb_hash_compare<typename strip<K>::type > >
+        struct key_matching {
+            typedef K key_type;
+            typedef typename strip<K>::type base_key_type;
+            typedef KHash hash_compare_type;
+        };
+
+        // old tag_matching join's new specifier
+        typedef key_matching<tag_value> tag_matching;
     }
 
 // -------------- function_body containers ----------------------
@@ -40,29 +62,20 @@ namespace internal {
         virtual ~source_body() {}
         virtual bool operator()(Output &output) = 0;
         virtual source_body* clone() = 0;
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        virtual void reset_body() = 0;
-#endif
     };
 
     //! The leaf for source_body
     template< typename Output, typename Body>
     class source_body_leaf : public source_body<Output> {
     public:
-        source_body_leaf( const Body &_body ) : body(_body), init_body(_body) { }
+        source_body_leaf( const Body &_body ) : body(_body) { }
         /*override*/ bool operator()(Output &output) { return body( output ); }
         /*override*/ source_body_leaf* clone() {
-            return new source_body_leaf< Output, Body >(init_body);
+            return new source_body_leaf< Output, Body >(body);
         }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        /*override*/ void reset_body() {
-            body = init_body;
-        }
-#endif
         Body get_body() { return body; }
     private:
         Body body;
-        Body init_body;
     };
 
     //! A functor that takes an Input and generates an Output
@@ -72,97 +85,70 @@ namespace internal {
         virtual ~function_body() {}
         virtual Output operator()(const Input &input) = 0;
         virtual function_body* clone() = 0;
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        virtual void reset_body() = 0;
-#endif
     };
 
     //! the leaf for function_body
     template <typename Input, typename Output, typename B>
     class function_body_leaf : public function_body< Input, Output > {
     public:
-        function_body_leaf( const B &_body ) : body(_body), init_body(_body) { }
+        function_body_leaf( const B &_body ) : body(_body) { }
         Output operator()(const Input &i) { return body(i); }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        /*override*/ void reset_body() {
-            body = init_body;
-        }
-#endif
         B get_body() { return body; }
         /*override*/ function_body_leaf* clone() {
-            return new function_body_leaf< Input, Output, B >(init_body);
+            return new function_body_leaf< Input, Output, B >(body);
         }
     private:
         B body;
-        B init_body;
     };
 
     //! the leaf for function_body specialized for Input and output of continue_msg
     template <typename B>
     class function_body_leaf< continue_msg, continue_msg, B> : public function_body< continue_msg, continue_msg > {
     public:
-        function_body_leaf( const B &_body ) : body(_body), init_body(_body) { }
+        function_body_leaf( const B &_body ) : body(_body) { }
         continue_msg operator()( const continue_msg &i ) {
             body(i);
             return i;
         }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        /*override*/ void reset_body() {
-            body = init_body;
-        }
-#endif
         B get_body() { return body; }
         /*override*/ function_body_leaf* clone() {
-           return new function_body_leaf< continue_msg, continue_msg, B >(init_body);
+           return new function_body_leaf< continue_msg, continue_msg, B >(body);
         }
     private:
         B body;
-        B init_body;
     };
 
     //! the leaf for function_body specialized for Output of continue_msg
     template <typename Input, typename B>
     class function_body_leaf< Input, continue_msg, B> : public function_body< Input, continue_msg > {
     public:
-        function_body_leaf( const B &_body ) : body(_body), init_body(_body) { }
+        function_body_leaf( const B &_body ) : body(_body) { }
         continue_msg operator()(const Input &i) {
             body(i);
             return continue_msg();
         }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        /*override*/ void reset_body() {
-            body = init_body;
-        }
-#endif
         B get_body() { return body; }
         /*override*/ function_body_leaf* clone() {
-            return new function_body_leaf< Input, continue_msg, B >(init_body);
+            return new function_body_leaf< Input, continue_msg, B >(body);
         }
     private:
         B body;
-        B init_body;
     };
 
     //! the leaf for function_body specialized for Input of continue_msg
     template <typename Output, typename B>
     class function_body_leaf< continue_msg, Output, B > : public function_body< continue_msg, Output > {
     public:
-        function_body_leaf( const B &_body ) : body(_body), init_body(_body) { }
+        function_body_leaf( const B &_body ) : body(_body) { }
         Output operator()(const continue_msg &i) {
             return body(i);
         }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        /*override*/ void reset_body() {
-            body = init_body;
-        }
-#endif
         B get_body() { return body; }
         /*override*/ function_body_leaf* clone() {
-            return new function_body_leaf< continue_msg, Output, B >(init_body);
+            return new function_body_leaf< continue_msg, Output, B >(body);
         }
     private:
         B body;
-        B init_body;
     };
 
     //! function_body that takes an Input and a set of output ports
@@ -172,33 +158,101 @@ namespace internal {
         virtual ~multifunction_body () {}
         virtual void operator()(const Input &/* input*/, OutputSet &/*oset*/) = 0;
         virtual multifunction_body* clone() = 0;
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        virtual void reset_body() = 0;
-#endif
     };
 
     //! leaf for multifunction.  OutputSet can be a std::tuple or a vector.
     template<typename Input, typename OutputSet, typename B>
     class multifunction_body_leaf : public multifunction_body<Input, OutputSet> {
     public:
-        multifunction_body_leaf(const B &_body) : body(_body), init_body(_body) { }
+        multifunction_body_leaf(const B &_body) : body(_body) { }
         void operator()(const Input &input, OutputSet &oset) {
             body(input, oset); // body may explicitly put() to one or more of oset.
         }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        /*override*/ void reset_body() {
-            body = init_body;
-        }
-#endif
         B get_body() { return body; }
         /*override*/ multifunction_body_leaf* clone() {
-            return new multifunction_body_leaf<Input, OutputSet,B>(init_body);
+            return new multifunction_body_leaf<Input, OutputSet,B>(body);
         }
     private:
         B body;
-        B init_body;
     };
 
+// ------ function bodies for hash_buffers and key-matching joins.
+
+template<typename Input, typename Output>
+class type_to_key_function_body : tbb::internal::no_assign {
+    public:
+        virtual ~type_to_key_function_body() {}
+        virtual Output operator()(const Input &input) = 0;  // returns an Output
+        virtual type_to_key_function_body* clone() = 0;
+};
+
+// specialization for ref output
+template<typename Input, typename Output>
+class type_to_key_function_body<Input,Output&> : tbb::internal::no_assign {
+    public:
+        virtual ~type_to_key_function_body() {}
+        virtual const Output & operator()(const Input &input) = 0;  // returns a const Output&
+        virtual type_to_key_function_body* clone() = 0;
+};
+
+template <typename Input, typename Output, typename B>
+class type_to_key_function_body_leaf : public type_to_key_function_body<Input, Output> {
+public:
+    type_to_key_function_body_leaf( const B &_body ) : body(_body) { }
+    /*override*/Output operator()(const Input &i) { return body(i); }
+    B get_body() { return body; }
+    /*override*/ type_to_key_function_body_leaf* clone() {
+        return new type_to_key_function_body_leaf< Input, Output, B>(body);
+    }
+private:
+    B body;
+};
+
+template <typename Input, typename Output, typename B>
+class type_to_key_function_body_leaf<Input,Output&,B> : public type_to_key_function_body< Input, Output&> {
+public:
+    type_to_key_function_body_leaf( const B &_body ) : body(_body) { }
+
+    /*override*/const Output& operator()(const Input &i) {
+        return const_cast<const Output&>(body(i));
+    }
+
+    B get_body() { return body; }
+
+    /*override*/ type_to_key_function_body_leaf* clone() {
+        return new type_to_key_function_body_leaf< Input, Output&, B>(body);
+    }
+
+private:
+    B body;
+};
+
+#if __TBB_PREVIEW_ASYNC_NODE
+
+    //! A functor that takes Input and submit it to Asynchronous Activity
+    template< typename Input, typename AsyncGateway >
+    class async_body : tbb::internal::no_assign {
+    public:
+        virtual ~async_body() {}
+        virtual void operator()(const Input &output, AsyncGateway& gateway) = 0;
+        virtual async_body* clone() = 0;
+    };
+
+    //! The leaf for async_body
+    template< typename Input, typename Body, typename AsyncGateway >
+    class async_body_leaf : public async_body< Input, AsyncGateway > {
+    public:
+        async_body_leaf( const Body &_body ) : body(_body) { }
+        /*override*/ void operator()(const Input &input, AsyncGateway& gateway) { body( input, gateway ); }
+        /*override*/ async_body_leaf* clone() {
+            return new async_body_leaf< Input, Body, AsyncGateway >(body);
+        }
+        Body get_body() { return body; }
+    private:
+        Body body;
+    };
+#endif
+
 // --------------------------- end of function_body containers ------------------------
 
 // --------------------------- node task bodies ---------------------------------------
@@ -272,17 +326,17 @@ namespace internal {
         typedef size_t size_type;
 
         bool empty() {
-            typename my_mutex_type::scoped_lock lock( my_mutex );
+            typename mutex_type::scoped_lock lock( my_mutex );
             return internal_empty();
         }
 
         void add( T &n ) {
-            typename my_mutex_type::scoped_lock lock( my_mutex );
+            typename mutex_type::scoped_lock lock( my_mutex );
             internal_push(n);
         }
 
         void remove( T &n ) {
-            typename my_mutex_type::scoped_lock lock( my_mutex );
+            typename mutex_type::scoped_lock lock( my_mutex );
             for ( size_t i = internal_size(); i != 0; --i ) {
                 T &s = internal_pop();
                 if ( &s == &n )  return;  // only remove one predecessor per request
@@ -290,36 +344,46 @@ namespace internal {
             }
         }
 
+        void clear() {
+            while( !my_q.empty()) (void)my_q.pop();
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        typedef std::vector<T *> predecessor_vector_type;
+            my_built_predecessors.clear();
+#endif
+        }
+
+#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+        typedef edge_container<T> built_predecessors_type;
+        built_predecessors_type &built_predecessors() { return my_built_predecessors; }
+
+        typedef typename edge_container<T>::edge_list_type predecessor_list_type;
         void internal_add_built_predecessor( T &n ) {
-            typename my_mutex_type::scoped_lock lock( my_mutex );
+            typename mutex_type::scoped_lock lock( my_mutex );
             my_built_predecessors.add_edge(n);
         }
 
         void internal_delete_built_predecessor( T &n ) {
-            typename my_mutex_type::scoped_lock lock( my_mutex );
+            typename mutex_type::scoped_lock lock( my_mutex );
             my_built_predecessors.delete_edge(n);
         }
 
-        void copy_predecessors( predecessor_vector_type &v) {
-            typename my_mutex_type::scoped_lock lock( my_mutex );
+        void copy_predecessors( predecessor_list_type &v) {
+            typename mutex_type::scoped_lock lock( my_mutex );
             my_built_predecessors.copy_edges(v);
         }
 
         size_t predecessor_count() {
-            typename my_mutex_type::scoped_lock lock(my_mutex);
+            typename mutex_type::scoped_lock lock(my_mutex);
             return (size_t)(my_built_predecessors.edge_count());
         }
-#endif  /* TBB_PREVIEW_FLOW_GRAPH_FEATURES */ 
+#endif  /* TBB_PREVIEW_FLOW_GRAPH_FEATURES */
 
     protected:
 
-        typedef M my_mutex_type;
-        my_mutex_type my_mutex;
+        typedef M mutex_type;
+        mutex_type my_mutex;
         std::queue< T * > my_q;
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        edge_container<T> my_built_predecessors;
+        built_predecessors_type my_built_predecessors;
 #endif
 
         // Assumes lock is held
@@ -350,7 +414,7 @@ namespace internal {
     template< typename T, typename M=spin_mutex >
     class predecessor_cache : public node_cache< sender<T>, M > {
     public:
-        typedef M my_mutex_type;
+        typedef M mutex_type;
         typedef T output_type;
         typedef sender<output_type> predecessor_type;
         typedef receiver<output_type> successor_type;
@@ -366,7 +430,7 @@ namespace internal {
             do {
                 predecessor_type *src;
                 {
-                    typename my_mutex_type::scoped_lock lock(this->my_mutex);
+                    typename mutex_type::scoped_lock lock(this->my_mutex);
                     if ( this->internal_empty() ) {
                         break;
                     }
@@ -388,7 +452,8 @@ namespace internal {
             return msg;
         }
 
-        void reset( __TBB_PFG_RESET_ARG(reset_flags f)) {
+        // If we are removing arcs (rf_clear_edges), call clear() rather than reset().
+        void reset() {
             if(my_owner) {
                 for(;;) {
                     predecessor_type *src;
@@ -396,14 +461,9 @@ namespace internal {
                         if(this->internal_empty()) break;
                         src = &this->internal_pop();
                     }
-                        src->register_successor( *my_owner);
+                    src->register_successor( *my_owner);
                 }
             }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-            if (f&rf_extract && my_owner) 
-                my_built_predecessors.receiver_extract(*my_owner);
-            __TBB_ASSERT(!(f&rf_extract) || this->internal_empty(), "predecessor cache not empty");
-#endif
         }
 
     protected:
@@ -418,7 +478,7 @@ namespace internal {
     template< typename T, typename M=spin_mutex >
     class reservable_predecessor_cache : public predecessor_cache< T, M > {
     public:
-        typedef M my_mutex_type;
+        typedef M mutex_type;
         typedef T output_type;
         typedef sender<T> predecessor_type;
         typedef receiver<T> successor_type;
@@ -431,7 +491,7 @@ namespace internal {
 
             do {
                 {
-                    typename my_mutex_type::scoped_lock lock(this->my_mutex);
+                    typename mutex_type::scoped_lock lock(this->my_mutex);
                     if ( reserved_src || this->internal_empty() )
                         return false;
 
@@ -442,7 +502,7 @@ namespace internal {
                 msg = reserved_src->try_reserve( v );
 
                 if (msg == false) {
-                    typename my_mutex_type::scoped_lock lock(this->my_mutex);
+                    typename mutex_type::scoped_lock lock(this->my_mutex);
                     // Relinquish ownership of the edge
                     reserved_src->register_successor( *this->my_owner );
                     reserved_src = NULL;
@@ -469,9 +529,14 @@ namespace internal {
             return true;
         }
 
-        void reset( __TBB_PFG_RESET_ARG(reset_flags f)) {
+        void reset( ) {
             reserved_src = NULL;
-            predecessor_cache<T,M>::reset(__TBB_PFG_RESET_ARG(f));
+            predecessor_cache<T,M>::reset( );
+        }
+
+        void clear() {
+            reserved_src = NULL;
+            predecessor_cache<T,M>::clear();
         }
 
     private:
@@ -484,45 +549,45 @@ namespace internal {
     class successor_cache : tbb::internal::no_copy {
     protected:
 
-        typedef M my_mutex_type;
-        my_mutex_type my_mutex;
+        typedef M mutex_type;
+        mutex_type my_mutex;
 
+        typedef receiver<T> successor_type;
         typedef receiver<T> *pointer_type;
-        typedef std::list< pointer_type > my_successors_type;
+        typedef std::list< pointer_type > successors_type;
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        edge_container<receiver<T> > my_built_successors;
+        edge_container<successor_type> my_built_successors;
 #endif
-        my_successors_type my_successors;
+        successors_type my_successors;
 
         sender<T> *my_owner;
 
     public:
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        typedef std::vector<pointer_type> successor_vector_type;
-        void internal_add_built_successor( receiver<T> &r) {
-            typename my_mutex_type::scoped_lock l(my_mutex, true);
+        typedef typename edge_container<successor_type>::edge_list_type successor_list_type;
+
+        edge_container<successor_type> &built_successors() { return my_built_successors; }
+
+        void internal_add_built_successor( successor_type &r) {
+            typename mutex_type::scoped_lock l(my_mutex, true);
             my_built_successors.add_edge( r );
         }
 
-        void internal_delete_built_successor( receiver<T> &r) {
-            typename my_mutex_type::scoped_lock l(my_mutex, true);
+        void internal_delete_built_successor( successor_type &r) {
+            typename mutex_type::scoped_lock l(my_mutex, true);
             my_built_successors.delete_edge(r);
         }
 
-        void copy_successors( successor_vector_type &v) {
-            typename my_mutex_type::scoped_lock l(my_mutex, false);
+        void copy_successors( successor_list_type &v) {
+            typename mutex_type::scoped_lock l(my_mutex, false);
             my_built_successors.copy_edges(v);
         }
 
         size_t successor_count() {
-            typename my_mutex_type::scoped_lock l(my_mutex,false);
+            typename mutex_type::scoped_lock l(my_mutex,false);
             return my_built_successors.edge_count();
         }
 
-        void reset( __TBB_PFG_RESET_ARG(reset_flags f)) {
-            if (f&rf_extract && my_owner) 
-                my_built_successors.sender_extract(*my_owner);
-        }
 #endif /* TBB_PREVIEW_FLOW_GRAPH_FEATURES */
 
         successor_cache( ) : my_owner(NULL) {}
@@ -531,14 +596,14 @@ namespace internal {
 
         virtual ~successor_cache() {}
 
-        void register_successor( receiver<T> &r ) {
-            typename my_mutex_type::scoped_lock l(my_mutex, true);
+        void register_successor( successor_type &r ) {
+            typename mutex_type::scoped_lock l(my_mutex, true);
             my_successors.push_back( &r );
         }
 
-        void remove_successor( receiver<T> &r ) {
-            typename my_mutex_type::scoped_lock l(my_mutex, true);
-            for ( typename my_successors_type::iterator i = my_successors.begin();
+        void remove_successor( successor_type &r ) {
+            typename mutex_type::scoped_lock l(my_mutex, true);
+            for ( typename successors_type::iterator i = my_successors.begin();
                   i != my_successors.end(); ++i ) {
                 if ( *i == & r ) {
                     my_successors.erase(i);
@@ -548,7 +613,7 @@ namespace internal {
         }
 
         bool empty() {
-            typename my_mutex_type::scoped_lock l(my_mutex, false);
+            typename mutex_type::scoped_lock l(my_mutex, false);
             return my_successors.empty();
         }
 
@@ -560,21 +625,23 @@ namespace internal {
         }
 
         virtual task * try_put_task( const T &t ) = 0;
-     };
+     };  // successor_cache<T>
 
     //! An abstract cache of successors, specialized to continue_msg
     template<>
     class successor_cache< continue_msg > : tbb::internal::no_copy {
     protected:
 
-        typedef spin_rw_mutex my_mutex_type;
-        my_mutex_type my_mutex;
+        typedef spin_rw_mutex mutex_type;
+        mutex_type my_mutex;
 
+        typedef receiver<continue_msg> successor_type;
         typedef receiver<continue_msg> *pointer_type;
-        typedef std::list< pointer_type > my_successors_type;
-        my_successors_type my_successors;
+        typedef std::list< pointer_type > successors_type;
+        successors_type my_successors;
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        edge_container<receiver<continue_msg> > my_built_successors;
+        edge_container<successor_type> my_built_successors;
+        typedef edge_container<successor_type>::edge_list_type successor_list_type;
 #endif
 
         sender<continue_msg> *my_owner;
@@ -582,31 +649,29 @@ namespace internal {
     public:
 
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-        typedef std::vector<pointer_type> successor_vector_type;
-        void internal_add_built_successor( receiver<continue_msg> &r) {
-            my_mutex_type::scoped_lock l(my_mutex, true);
+
+        edge_container<successor_type> &built_successors() { return my_built_successors; }
+
+        void internal_add_built_successor( successor_type &r) {
+            mutex_type::scoped_lock l(my_mutex, true);
             my_built_successors.add_edge( r );
         }
 
-        void internal_delete_built_successor( receiver<continue_msg> &r) {
-            my_mutex_type::scoped_lock l(my_mutex, true);
+        void internal_delete_built_successor( successor_type &r) {
+            mutex_type::scoped_lock l(my_mutex, true);
             my_built_successors.delete_edge(r);
         }
 
-        void copy_successors( successor_vector_type &v) {
-            my_mutex_type::scoped_lock l(my_mutex, false);
+        void copy_successors( successor_list_type &v) {
+            mutex_type::scoped_lock l(my_mutex, false);
             my_built_successors.copy_edges(v);
         }
 
         size_t successor_count() {
-            my_mutex_type::scoped_lock l(my_mutex,false);
+            mutex_type::scoped_lock l(my_mutex,false);
             return my_built_successors.edge_count();
         }
 
-        void reset( __TBB_PFG_RESET_ARG(reset_flags f)) {
-            if (f&rf_extract && my_owner) 
-                my_built_successors.sender_extract(*my_owner);
-        }
 #endif  /* TBB_PREVIEW_FLOW_GRAPH_FEATURES */
 
         successor_cache( ) : my_owner(NULL) {}
@@ -615,17 +680,17 @@ namespace internal {
 
         virtual ~successor_cache() {}
 
-        void register_successor( receiver<continue_msg> &r ) {
-            my_mutex_type::scoped_lock l(my_mutex, true);
+        void register_successor( successor_type &r ) {
+            mutex_type::scoped_lock l(my_mutex, true);
             my_successors.push_back( &r );
             if ( my_owner && r.is_continue_receiver() ) {
                 r.register_predecessor( *my_owner );
             }
         }
 
-        void remove_successor( receiver<continue_msg> &r ) {
-            my_mutex_type::scoped_lock l(my_mutex, true);
-            for ( my_successors_type::iterator i = my_successors.begin();
+        void remove_successor( successor_type &r ) {
+            mutex_type::scoped_lock l(my_mutex, true);
+            for ( successors_type::iterator i = my_successors.begin();
                   i != my_successors.end(); ++i ) {
                 if ( *i == & r ) {
                     // TODO: Check if we need to test for continue_receiver before
@@ -639,7 +704,7 @@ namespace internal {
         }
 
         bool empty() {
-            my_mutex_type::scoped_lock l(my_mutex, false);
+            mutex_type::scoped_lock l(my_mutex, false);
             return my_successors.empty();
         }
 
@@ -652,13 +717,13 @@ namespace internal {
 
         virtual task * try_put_task( const continue_msg &t ) = 0;
 
-     };
+    };  // successor_cache< continue_msg >
 
     //! A cache of successors that are broadcast to
     template<typename T, typename M=spin_rw_mutex>
     class broadcast_cache : public successor_cache<T, M> {
-        typedef M my_mutex_type;
-        typedef std::list< receiver<T> * > my_successors_type;
+        typedef M mutex_type;
+        typedef typename successor_cache<T,M>::successors_type successors_type;
 
     public:
 
@@ -668,8 +733,8 @@ namespace internal {
         /*override*/ task * try_put_task( const T &t ) {
             task * last_task = NULL;
             bool upgraded = true;
-            typename my_mutex_type::scoped_lock l(this->my_mutex, upgraded);
-            typename my_successors_type::iterator i = this->my_successors.begin();
+            typename mutex_type::scoped_lock l(this->my_mutex, upgraded);
+            typename successors_type::iterator i = this->my_successors.begin();
             while ( i != this->my_successors.end() ) {
                 task *new_task = (*i)->try_put_task(t);
                 last_task = combine_tasks(last_task, new_task);  // enqueue if necessary
@@ -697,22 +762,22 @@ namespace internal {
     template<typename T, typename M=spin_rw_mutex >
     class round_robin_cache : public successor_cache<T, M> {
         typedef size_t size_type;
-        typedef M my_mutex_type;
-        typedef std::list< receiver<T> * > my_successors_type;
+        typedef M mutex_type;
+        typedef typename successor_cache<T,M>::successors_type successors_type;
 
     public:
 
         round_robin_cache( ) {}
 
         size_type size() {
-            typename my_mutex_type::scoped_lock l(this->my_mutex, false);
+            typename mutex_type::scoped_lock l(this->my_mutex, false);
             return this->my_successors.size();
         }
 
         /*override*/task *try_put_task( const T &t ) {
             bool upgraded = true;
-            typename my_mutex_type::scoped_lock l(this->my_mutex, upgraded);
-            typename my_successors_type::iterator i = this->my_successors.begin();
+            typename mutex_type::scoped_lock l(this->my_mutex, upgraded);
+            typename successors_type::iterator i = this->my_successors.begin();
             while ( i != this->my_successors.end() ) {
                 task *new_task = (*i)->try_put_task(t);
                 if ( new_task ) {
