@@ -58,7 +58,7 @@ namespace internal {
             blt_pred_cnt, blt_pred_cpy   // create vector copies of preds and succs
 #endif 
         };
-        typedef function_input_base<Input, A, ImplType> my_class;
+        typedef function_input_base<Input, A, ImplType> class_type;
         
     public:
 
@@ -79,7 +79,7 @@ namespace internal {
             : my_graph(g), my_max_concurrency(max_concurrency), my_concurrency(0),
               my_queue(q), forwarder_busy(false) {
             my_predecessors.set_owner(this);
-            my_aggregator.initialize_handler(my_handler(this));
+            my_aggregator.initialize_handler(handler_type(this));
         }
         
         //! Copy constructor
@@ -89,7 +89,7 @@ namespace internal {
             my_concurrency(0), my_queue(q), forwarder_busy(false)
         {
             my_predecessors.set_owner(this);
-            my_aggregator.initialize_handler(my_handler(this));
+            my_aggregator.initialize_handler(handler_type(this));
         }
 
         //! Destructor
@@ -105,7 +105,7 @@ namespace internal {
            if ( my_max_concurrency == 0 ) {
                return create_body_task( t );
            } else {
-               my_operation op_data(t, tryput_bypass);
+               operation_type op_data(t, tryput_bypass);
                my_aggregator.execute(&op_data);
                if(op_data.status == SUCCEEDED ) {
                    return op_data.bypass_t;
@@ -116,7 +116,7 @@ namespace internal {
 
         //! Adds src to the list of cached predecessors.
         /* override */ bool register_predecessor( predecessor_type &src ) {
-            my_operation op_data(reg_pred);
+            operation_type op_data(reg_pred);
             op_data.r = &src;
             my_aggregator.execute(&op_data);
             return true;
@@ -124,7 +124,7 @@ namespace internal {
         
         //! Removes src from the list of cached predecessors.
         /* override */ bool remove_predecessor( predecessor_type &src ) {
-            my_operation op_data(rem_pred);
+            operation_type op_data(rem_pred);
             op_data.r = &src;
             my_aggregator.execute(&op_data);
             return true;
@@ -133,26 +133,26 @@ namespace internal {
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
         //! Adds to list of predecessors added by make_edge
         /*override*/ void internal_add_built_predecessor( predecessor_type &src) {
-            my_operation op_data(add_blt_pred);
+            operation_type op_data(add_blt_pred);
             op_data.r = &src;
             my_aggregator.execute(&op_data);
         }
 
         //! removes from to list of predecessors (used by remove_edge)
         /*override*/ void internal_delete_built_predecessor( predecessor_type &src) {
-            my_operation op_data(del_blt_pred);
+            operation_type op_data(del_blt_pred);
             op_data.r = &src;
             my_aggregator.execute(&op_data);
         }
 
         /*override*/ size_t predecessor_count() {
-            my_operation op_data(blt_pred_cnt);
+            operation_type op_data(blt_pred_cnt);
             my_aggregator.execute(&op_data);
             return op_data.cnt_val;
         }
 
         /*override*/ void copy_predecessors(predecessor_list_type &v) {
-            my_operation op_data(blt_pred_cpy);
+            operation_type op_data(blt_pred_cpy);
             op_data.predv = &v;
             my_aggregator.execute(&op_data);
         }
@@ -164,12 +164,12 @@ namespace internal {
 
     protected:
 
-        void reset_function_input_base( __TBB_PFG_RESET_ARG(reset_flags f)) {
+        void reset_function_input_base( reset_flags f) {
             my_concurrency = 0;
             if(my_queue) {
                 my_queue->reset();
             }
-            reset_receiver(__TBB_PFG_RESET_ARG(f));
+            reset_receiver(f);
             forwarder_busy = false;
         }
 
@@ -179,23 +179,19 @@ namespace internal {
         input_queue_type *my_queue;
         predecessor_cache<input_type, null_mutex > my_predecessors;
         
-        /*override*/void reset_receiver( __TBB_PFG_RESET_ARG(reset_flags f)) {
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+        /*override*/void reset_receiver( reset_flags f) {
             if( f & rf_clear_edges) my_predecessors.clear();
             else
-#endif
                 my_predecessors.reset();
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
             __TBB_ASSERT(!(f & rf_clear_edges) || my_predecessors.empty(), "function_input_base reset failed");
-#endif
         }
 
     private:
 
-        friend class apply_body_task_bypass< my_class, input_type >;
-        friend class forward_task_bypass< my_class >;
+        friend class apply_body_task_bypass< class_type, input_type >;
+        friend class forward_task_bypass< class_type >;
         
-        class my_operation : public aggregated_operation< my_operation > {
+        class operation_type : public aggregated_operation< operation_type > {
         public:
             char type;
             union {
@@ -207,18 +203,18 @@ namespace internal {
 #endif  /* TBB_PREVIEW_FLOW_GRAPH_FEATURES */
             };
             tbb::task *bypass_t;
-            my_operation(const input_type& e, op_type t) :
+            operation_type(const input_type& e, op_type t) :
                 type(char(t)), elem(const_cast<input_type*>(&e)) {}
-            my_operation(op_type t) : type(char(t)), r(NULL) {}
+            operation_type(op_type t) : type(char(t)), r(NULL) {}
         };
         
         bool forwarder_busy;
-        typedef internal::aggregating_functor<my_class, my_operation> my_handler;
-        friend class internal::aggregating_functor<my_class, my_operation>;
-        aggregator< my_handler, my_operation > my_aggregator;
+        typedef internal::aggregating_functor<class_type, operation_type> handler_type;
+        friend class internal::aggregating_functor<class_type, operation_type>;
+        aggregator< handler_type, operation_type > my_aggregator;
         
-        void handle_operations(my_operation *op_list) {
-            my_operation *tmp;
+        void handle_operations(operation_type *op_list) {
+            operation_type *tmp;
             while (op_list) {
                 tmp = op_list;
                 op_list = op_list->next;
@@ -298,7 +294,7 @@ namespace internal {
         }
         
         //! Put to the node, but return the task instead of enqueueing it
-        void internal_try_put_task(my_operation *op) {
+        void internal_try_put_task(operation_type *op) {
             __TBB_ASSERT(my_max_concurrency != 0, NULL);
             if (my_concurrency < my_max_concurrency) {
                ++my_concurrency;
@@ -315,7 +311,7 @@ namespace internal {
         }
         
         //! Tries to spawn bodies if available and if concurrency allows
-        void internal_forward(my_operation *op) {
+        void internal_forward(operation_type *op) {
             op->bypass_t = NULL;
             if (my_concurrency<my_max_concurrency || !my_max_concurrency) {
                 input_type i;
@@ -350,7 +346,7 @@ namespace internal {
         task * apply_body_bypass( input_type &i ) {
             task * new_task = static_cast<ImplType *>(this)->apply_body_impl_bypass(i);
             if ( my_max_concurrency != 0 ) {
-                my_operation op_data(app_body_bypass);  // tries to pop an item or get_item, enqueues another apply_body
+                operation_type op_data(app_body_bypass);  // tries to pop an item or get_item, enqueues another apply_body
                 my_aggregator.execute(&op_data);
                 tbb::task *ttask = op_data.bypass_t;
                 new_task = combine_tasks(new_task, ttask);
@@ -361,10 +357,9 @@ namespace internal {
         //! allocates a task to call apply_body( input )
         inline task * create_body_task( const input_type &input ) {
             
-            task* tp = my_graph.root_task();
-            return (tp) ?
-                new(task::allocate_additional_child_of(*tp))
-                    apply_body_task_bypass < my_class, input_type >(*this, input) :
+            return (my_graph.is_active()) ?
+                new(task::allocate_additional_child_of(*(my_graph.root_task())))
+                    apply_body_task_bypass < class_type, input_type >(*this, input) :
                 NULL;
         }
 
@@ -379,7 +374,7 @@ namespace internal {
         
        //! This is executed by an enqueued task, the "forwarder"
        task *forward_task() {
-           my_operation op_data(try_fwd);
+           operation_type op_data(try_fwd);
            task *rval = NULL;
            do {
                op_data.status = WAIT;
@@ -393,9 +388,8 @@ namespace internal {
        }
         
        inline task *create_forward_task() {
-           task* tp = my_graph.root_task();
-           return (tp) ?
-               new(task::allocate_additional_child_of(*tp)) forward_task_bypass< my_class >(*this) :
+           return (my_graph.is_active()) ?
+               new(task::allocate_additional_child_of(*(my_graph.root_task()))) forward_task_bypass< class_type >(*this) :
                NULL;
        }
 
@@ -415,6 +409,7 @@ namespace internal {
     public:
         typedef Input input_type;
         typedef Output output_type;
+        typedef function_body<input_type, output_type> function_body_type;
         typedef function_input<Input,Output,A> my_class;
         typedef function_input_base<Input, A, my_class> base_type;
         typedef function_input_queue<input_type, A> input_queue_type;
@@ -423,22 +418,25 @@ namespace internal {
         template<typename Body>
         function_input( graph &g, size_t max_concurrency, Body& body, input_queue_type *q = NULL ) :
             base_type(g, max_concurrency, q),
-            my_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ) {
+            my_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ),
+            my_init_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ) {
         }
 
         //! Copy constructor
         function_input( const function_input& src, input_queue_type *q = NULL ) : 
                 base_type(src, q),
-                my_body( src.my_body->clone() ) {
+                my_body( src.my_init_body->clone() ),
+                my_init_body(src.my_init_body->clone() ) {
         }
 
         ~function_input() {
             delete my_body;
+            delete my_init_body;
         }
 
         template< typename Body >
         Body copy_function_object() {
-            internal::function_body<input_type, output_type> &body_ref = *this->my_body;
+            function_body_type &body_ref = *this->my_body;
             return dynamic_cast< internal::function_body_leaf<input_type, output_type, Body> & >(body_ref).get_body(); 
         } 
 
@@ -458,20 +456,22 @@ namespace internal {
 
     protected:
 
-        void reset_function_input(__TBB_PFG_RESET_ARG(reset_flags f)) {
-            base_type::reset_function_input_base(__TBB_PFG_RESET_ARG(f));
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-            if(f & rf_reset_bodies) my_body->reset_body();
-#endif
+        void reset_function_input(reset_flags f) {
+            base_type::reset_function_input_base(f);
+            if(f & rf_reset_bodies) {
+                function_body_type *tmp = my_init_body->clone();
+                delete my_body;
+                my_body = tmp;
+            }
         }
 
-        function_body<input_type, output_type> *my_body;
+        function_body_type *my_body;
+        function_body_type *my_init_body;
         virtual broadcast_cache<output_type > &successors() = 0;
 
     };  // function_input
 
 
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
     // helper templates to clear the successor edges of the output ports of an multifunction_node
     template<int N> struct clear_element {
         template<typename P> static void clear_this(P &p) {
@@ -494,6 +494,7 @@ namespace internal {
         }
     };
 
+#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
     // helper templates to extract the output ports of an multifunction_node from graph
     template<int N> struct extract_element {
         template<typename P> static void extract_this(P &p) {
@@ -517,6 +518,7 @@ namespace internal {
         static const int N = tbb::flow::tuple_size<OutputPortSet>::value;
         typedef Input input_type;
         typedef OutputPortSet output_ports_type;
+        typedef multifunction_body<input_type, output_ports_type> multifunction_body_type;
         typedef multifunction_input<Input,OutputPortSet,A> my_class;
         typedef function_input_base<Input, A, my_class> base_type;
         typedef function_input_queue<input_type, A> input_queue_type;
@@ -529,22 +531,25 @@ namespace internal {
                 Body& body,
                 input_queue_type *q = NULL ) :
             base_type(g, max_concurrency, q),
-            my_body( new internal::multifunction_body_leaf<input_type, output_ports_type, Body>(body) ) {
+            my_body( new internal::multifunction_body_leaf<input_type, output_ports_type, Body>(body) ),
+            my_init_body( new internal::multifunction_body_leaf<input_type, output_ports_type, Body>(body) ) {
         }
 
         //! Copy constructor
         multifunction_input( const multifunction_input& src, input_queue_type *q = NULL ) : 
                 base_type(src, q),
-                my_body( src.my_body->clone() ) {
+                my_body( src.my_init_body->clone() ),
+                my_init_body(src.my_init_body->clone() ) {
         }
 
         ~multifunction_input() {
             delete my_body;
+            delete my_init_body;
         }
 
         template< typename Body >
         Body copy_function_object() {
-            internal::multifunction_body<input_type, output_ports_type> &body_ref = *this->my_body;
+            multifunction_body_type &body_ref = *this->my_body;
             return dynamic_cast< internal::multifunction_body_leaf<input_type, output_ports_type, Body> & >(body_ref).get_body(); 
         } 
 
@@ -567,16 +572,19 @@ namespace internal {
         }
 #endif
 
-        /*override*/void reset(__TBB_PFG_RESET_ARG(reset_flags f)) { 
-            base_type::reset_function_input_base(__TBB_PFG_RESET_ARG(f));
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+        /*override*/void reset(reset_flags f) { 
+            base_type::reset_function_input_base(f);
             if(f & rf_clear_edges)clear_element<N>::clear_this(my_output_ports);
-            if(f & rf_reset_bodies) my_body->reset_body();
+            if(f & rf_reset_bodies) {
+                multifunction_body_type *tmp = my_init_body->clone();
+                delete my_body;
+                my_body = tmp;
+            }
             __TBB_ASSERT(!(f & rf_clear_edges) || clear_element<N>::this_empty(my_output_ports), "multifunction_node reset failed");
-#endif
         }
 
-        multifunction_body<input_type, output_ports_type> *my_body;
+        multifunction_body_type *my_body;
+        multifunction_body_type *my_init_body;
         output_ports_type my_output_ports;
 
     };  // multifunction_input
@@ -615,41 +623,51 @@ namespace internal {
             
         //! The output type of this receiver
         typedef Output output_type;
+        typedef function_body<input_type, output_type> function_body_type;
         
         template< typename Body >
         continue_input( graph &g, Body& body )
             : my_graph_ptr(&g), 
-             my_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ) { }
+             my_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ),
+             my_init_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ) { }
         
         template< typename Body >
         continue_input( graph &g, int number_of_predecessors, Body& body )
             : continue_receiver( number_of_predecessors ), my_graph_ptr(&g), 
-             my_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ) { }
+             my_body( new internal::function_body_leaf< input_type, output_type, Body>(body) ),
+             my_init_body( new internal::function_body_leaf< input_type, output_type, Body>(body) )
+        { }
 
         continue_input( const continue_input& src ) : continue_receiver(src), 
-            my_graph_ptr(src.my_graph_ptr), my_body( src.my_body->clone() ) {}
+            my_graph_ptr(src.my_graph_ptr),
+            my_body( src.my_init_body->clone() ),
+            my_init_body( src.my_init_body->clone() ) {}
 
         ~continue_input() {
             delete my_body;
+            delete my_init_body;
         }
 
         template< typename Body >
         Body copy_function_object() {
-            internal::function_body<input_type, output_type> &body_ref = *my_body;
+            function_body_type &body_ref = *my_body;
             return dynamic_cast< internal::function_body_leaf<input_type, output_type, Body> & >(body_ref).get_body(); 
         } 
 
-        /*override*/void reset_receiver( __TBB_PFG_RESET_ARG(reset_flags f)) {
-            continue_receiver::reset_receiver(__TBB_PFG_RESET_ARG(f));
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-            if(f & rf_reset_bodies) my_body->reset_body();
-#endif
+        /*override*/void reset_receiver( reset_flags f) {
+            continue_receiver::reset_receiver(f);
+            if(f & rf_reset_bodies) {
+                function_body_type *tmp = my_init_body->clone();
+                delete my_body;
+                my_body = tmp;
+            }
         }
 
     protected:
         
         graph* my_graph_ptr;
-        function_body<input_type, output_type> *my_body;
+        function_body_type *my_body;
+        function_body_type *my_init_body;
         
         virtual broadcast_cache<output_type > &successors() = 0; 
         
@@ -671,9 +689,8 @@ namespace internal {
         
         //! Spawns a task that applies the body
         /* override */ task *execute( ) {
-            task* tp = my_graph_ptr->root_task();
-            return (tp) ?
-                new ( task::allocate_additional_child_of( *tp ) ) 
+            return (my_graph_ptr->is_active()) ?
+                new ( task::allocate_additional_child_of( *(my_graph_ptr->root_task()) ) ) 
                     apply_body_task_bypass< continue_input< Output >, continue_msg >( *this, continue_msg() ) :
                 NULL;
         }
@@ -689,6 +706,7 @@ namespace internal {
     public:
         typedef Input input_type;
         typedef AsyncGatewayType async_gateway_type;
+        typedef async_body< input_type, async_gateway_type > async_body_type;
         typedef async_input< Input, A, async_gateway_type > my_class;
         typedef function_input_base<Input, A, my_class> base_type;
 
@@ -696,22 +714,26 @@ namespace internal {
         template<typename Body>
         async_input( graph &g, Body& body ) : 
             base_type( g, unlimited ),
-            my_body( new internal::async_body_leaf< input_type, Body, async_gateway_type >(body) ){
+            my_body( new internal::async_body_leaf< input_type, Body, async_gateway_type >(body) ),
+            my_init_body( new internal::async_body_leaf< input_type, Body, async_gateway_type >(body) )
+        {
         }
 
         //! Copy constructor
         async_input( const async_input& src ) : 
             base_type( src ),
-            my_body( src.my_body->clone() ) {
+            my_body( src.my_init_body->clone() ),
+            my_init_body(src.my_init_body->clone() ) {
         }
 
         ~async_input() {
             delete my_body;
+            delete my_init_body;
         }
 
         template< typename Body >
         Body copy_function_object() {
-            internal::async_body<input_type, async_gateway_type> &body_ref = *this->my_body;
+            async_body_type &body_ref = *this->my_body;
             return dynamic_cast< internal::async_body_leaf<input_type, Body, async_gateway_type> & >(body_ref).get_body(); 
         }
 
@@ -728,14 +750,17 @@ namespace internal {
         virtual async_gateway_type& async_gateway() = 0;
 
     protected:
-        void reset_async_input(__TBB_PFG_RESET_ARG(reset_flags f)) {
-            base_type::reset_function_input_base(__TBB_PFG_RESET_ARG(f));
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
-            if(f & rf_reset_bodies) my_body->reset_body();
-#endif
+        void reset_async_input(reset_flags f) {
+            base_type::reset_function_input_base(f);
+            if(f & rf_reset_bodies) {
+                async_body_type *tmp  = my_init_body->clone();
+                delete my_body;
+                my_body = tmp;
+            }
         }
 
-        async_body< input_type, async_gateway_type > *my_body;
+        async_body_type *my_body;
+        async_body_type *my_init_body;
     };
 #endif // __TBB_PREVIEW_ASYNC_NODE
         
@@ -744,9 +769,7 @@ namespace internal {
     class function_output : public sender<Output> {
     public:
         
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
         template<int N> friend struct clear_element;
-#endif
         typedef Output output_type;
         typedef receiver<output_type> successor_type;
         typedef broadcast_cache<output_type> broadcast_cache_type;
@@ -827,6 +850,23 @@ namespace internal {
         }
     };  // multifunction_output
 
+//composite_node
+#if TBB_PREVIEW_FLOW_GRAPH_TRACE
+    template<typename CompositeType>
+    void add_nodes_impl(CompositeType*, bool) {}
+
+    template< typename CompositeType, typename NodeType1, typename... NodeTypes >
+    void add_nodes_impl(CompositeType *c_node, bool visible, const NodeType1& n1, const NodeTypes&... n) {
+        void *addr = const_cast<NodeType1 *>(&n1);
+
+        if(visible)
+            tbb::internal::itt_relation_add( tbb::internal::ITT_DOMAIN_FLOW, c_node, tbb::internal::FLOW_NODE, tbb::internal::__itt_relation_is_parent_of, addr, tbb::internal::FLOW_NODE );
+        else
+            tbb::internal::itt_relation_add( tbb::internal::ITT_DOMAIN_FLOW, addr, tbb::internal::FLOW_NODE, tbb::internal::__itt_relation_is_child_of, c_node, tbb::internal::FLOW_NODE );
+        add_nodes_impl(c_node, visible, n...);
+    }
+#endif
+
 }  // internal
 
 #endif // __TBB__flow_graph_node_impl_H
