@@ -82,10 +82,6 @@ struct arena_base : padded<intrusive_list_node> {
     //! Number of workers requested by the master thread owning the arena
     unsigned my_max_num_workers;
 
-#if __TBB_TRACK_PRIORITY_LEVEL_SATURATION
-    int my_num_workers_present;
-#endif /* __TBB_TRACK_PRIORITY_LEVEL_SATURATION */
-
     //! Current task pool state and estimate of available tasks amount.
     /** The estimate is either 0 (SNAPSHOT_EMPTY) or infinity (SNAPSHOT_FULL). 
         Special state is "busy" (any other unsigned value). 
@@ -157,53 +153,25 @@ struct arena_base : padded<intrusive_list_node> {
 #endif /* TBB_USE_ASSERT */
 }; // struct arena_base
 
-class arena
-#if (__GNUC__<4 || __GNUC__==4 && __GNUC_MINOR__==0) && !__INTEL_COMPILER
-    : public padded<arena_base>
-#else
-    : private padded<arena_base>
-#endif
+class arena: public padded<arena_base>
 {
-private:
-    friend class generic_scheduler;
-    template<typename SchedulerTraits> friend class custom_scheduler;
-    friend class governor;
-    friend class task_scheduler_observer_v3;
-    friend class market;
-    friend class tbb::task;
-    friend class tbb::task_group_context;
-    friend class allocate_root_with_context_proxy;
-    friend class intrusive_list<arena>;
-    friend class interface7::internal::task_arena_base; // declared in scheduler_common.h
-    friend class interface7::internal::delegated_task;
-    friend class interface7::internal::wait_task;
-
+public:
     typedef padded<arena_base> base_type;
 
     //! Constructor
     arena ( market&, unsigned max_num_workers );
 
     //! Allocate an instance of arena.
-    static arena& allocate_arena( market&, unsigned max_num_workers );
+    static arena& allocate_arena( market&, unsigned num_slots );
 
-    static int unsigned num_slots_to_reserve ( unsigned max_num_workers ) {
-        return max(2u, max_num_workers + 1);
+    static int unsigned num_slots_to_reserve ( unsigned num_slots ) {
+        return max(2u, num_slots);
     }
 
     static int allocation_size ( unsigned max_num_workers ) {
         return sizeof(base_type) + num_slots_to_reserve(max_num_workers) * (sizeof(mail_outbox) + sizeof(arena_slot));
     }
 
-#if __TBB_TASK_GROUP_CONTEXT
-    //! Finds all contexts affected by the state change and propagates the new state to them.
-    /** The propagation is relayed to the market because tasks created by one 
-        master thread can be passed to and executed by other masters. This means 
-        that context trees can span several arenas at once and thus state change
-        propagation cannot be generally localized to one arena only. **/
-    template <typename T>
-    bool propagate_task_group_state ( T task_group_context::*mptr_state, task_group_context& src, T new_state );
-#endif /* __TBB_TASK_GROUP_CONTEXT */
-
     //! Get reference to mailbox corresponding to given affinity_id.
     mail_outbox& mailbox( affinity_id id ) {
         __TBB_ASSERT( 0<id, "affinity id must be positive integer" );
@@ -271,7 +239,6 @@ private:
     arena_slot my_slots[1];
 }; // class arena
 
-
 template<bool is_master>
 inline void arena::on_thread_leaving ( ) {
     //
@@ -314,7 +281,7 @@ inline void arena::on_thread_leaving ( ) {
     // In both cases we cannot dereference arena pointer after the refcount is
     // decremented, as our arena may already be destroyed.
     //
-    // If this is the master thread, market can be concurrently destroyed.
+    // If this is the master thread, the market is protected by refcount to it.
     // In case of workers market's liveness is ensured by the RML connection
     // rundown protocol, according to which the client (i.e. the market) lives
     // until RML server notifies it about connection termination, and this
@@ -327,7 +294,7 @@ inline void arena::on_thread_leaving ( ) {
     market* m = my_market;
     __TBB_ASSERT(my_references > int(!is_master), "broken arena reference counter");
     if ( (my_references -= is_master? 1:2 ) == 0 ) // worker's counter starts from bit 1
-        market::try_destroy_arena( m, this, aba_epoch, is_master );
+        m->try_destroy_arena( this, aba_epoch );
 }
 
 template<bool Spawned> void arena::advertise_new_work() {
