@@ -1,31 +1,25 @@
 /*
-    Copyright 2005-2012 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
 
-    This file is part of Threading Building Blocks.
+    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
+    you can redistribute it and/or modify it under the terms of the GNU General Public License
+    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
+    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
+    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+    See  the GNU General Public License for more details.   You should have received a copy of
+    the  GNU General Public License along with Threading Building Blocks; if not, write to the
+    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
 
-    Threading Building Blocks is free software; you can redistribute it
-    and/or modify it under the terms of the GNU General Public License
-    version 2 as published by the Free Software Foundation.
-
-    Threading Building Blocks is distributed in the hope that it will be
-    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with Threading Building Blocks; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-
-    As a special exception, you may use this file as part of a free software
-    library without restriction.  Specifically, if other files instantiate
-    templates or use macros or inline functions from this file, or you compile
-    this file and link it with other files to produce an executable, this
-    file does not by itself cause the resulting executable to be covered by
-    the GNU General Public License.  This exception does not however
-    invalidate any other reasons why the executable file might be covered by
-    the GNU General Public License.
+    As a special exception,  you may use this file  as part of a free software library without
+    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
+    functions from this file, or you compile this file and link it with other files to produce
+    an executable,  this file does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however invalidate any other
+    reasons why the executable file might be covered by the GNU General Public License.
 */
 
+#include <stdio.h>
+#include <stdlib.h>
 #include "governor.h"
 #include "tbb_main.h"
 #include "scheduler.h"
@@ -44,7 +38,7 @@ namespace internal {
 //------------------------------------------------------------------------
 
 #if __TBB_SURVIVE_THREAD_SWITCH
-// Support for interoperability with Intel(R) Cilk(tm) Plus.
+// Support for interoperability with Intel(R) Cilk(TM) Plus.
 
 #if _WIN32
 #define CILKLIB_NAME "cilkrts20.dll"
@@ -56,13 +50,9 @@ namespace internal {
 static __cilk_tbb_retcode (*watch_stack_handler)(struct __cilk_tbb_unwatch_thunk* u,
                                                  struct __cilk_tbb_stack_op_thunk o);
 
-#if __TBB_WEAK_SYMBOLS
-    #pragma weak __cilkrts_watch_stack
-#endif
-
 //! Table describing how to link the handlers.
 static const dynamic_link_descriptor CilkLinkTable[] = {
-    DLD(__cilkrts_watch_stack, watch_stack_handler)
+    { "__cilkrts_watch_stack", (pointer_to_handler*)(void*)(&watch_stack_handler) }
 };
 
 static atomic<do_once_state> cilkrts_load_state;
@@ -70,7 +60,7 @@ static atomic<do_once_state> cilkrts_load_state;
 bool initialize_cilk_interop() {
     // Pinning can fail. This is a normal situation, and means that the current
     // thread does not use cilkrts and consequently does not need interop.
-    return dynamic_link( CILKLIB_NAME, CilkLinkTable, 1 );
+    return dynamic_link( CILKLIB_NAME, CilkLinkTable, 1,  /*handle=*/0, DYNAMIC_LINK_GLOBAL );
 }
 #endif /* __TBB_SURVIVE_THREAD_SWITCH */
 
@@ -86,6 +76,7 @@ void governor::acquire_resources () {
 #endif
     if( status )
         handle_perror(status, "TBB failed to initialize task scheduler TLS\n");
+    is_speculation_enabled = cpu_has_speculation();
 }
 
 void governor::release_resources () {
@@ -96,7 +87,7 @@ void governor::release_resources () {
 #endif
     int status = theTLS.destroy();
     if( status )
-        handle_perror(status, "TBB failed to destroy task scheduler TLS");
+        runtime_warning("failed to destroy task scheduler TLS: %s", strerror(status));
     dynamic_unlink_all();
 }
 
@@ -117,9 +108,24 @@ rml::tbb_server* governor::create_rml_server ( rml::tbb_client& client ) {
     return server;
 }
 
+
+uintptr_t governor::tls_value_of( generic_scheduler* s ) {
+    __TBB_ASSERT( (uintptr_t(s)&1) == 0, "Bad pointer to the scheduler" );
+    // LSB marks the scheduler initialized with arena
+    return uintptr_t(s) | uintptr_t((s && (s->my_arena || s->is_worker()))? 1 : 0);
+}
+
+void governor::assume_scheduler( generic_scheduler* s ) {
+    theTLS.set( tls_value_of(s) );
+}
+
+bool governor::is_set( generic_scheduler* s ) {
+    return theTLS.get() == tls_value_of(s);
+}
+
 void governor::sign_on(generic_scheduler* s) {
-    __TBB_ASSERT( !theTLS.get(), NULL );
-    theTLS.set(s);
+    __TBB_ASSERT( is_set(NULL) && s, NULL );
+    assume_scheduler( s );
 #if __TBB_SURVIVE_THREAD_SWITCH
     if( watch_stack_handler ) {
         __cilk_tbb_stack_op_thunk o;
@@ -135,12 +141,13 @@ void governor::sign_on(generic_scheduler* s) {
 #endif /* TBB_USE_ASSERT */
     }
 #endif /* __TBB_SURVIVE_THREAD_SWITCH */
+    __TBB_ASSERT( is_set(s), NULL );
 }
 
 void governor::sign_off(generic_scheduler* s) {
     suppress_unused_warning(s);
-    __TBB_ASSERT( theTLS.get()==s, "attempt to unregister a wrong scheduler instance" );
-    theTLS.set(NULL);
+    __TBB_ASSERT( is_set(s), "attempt to unregister a wrong scheduler instance" );
+    assume_scheduler(NULL);
 #if __TBB_SURVIVE_THREAD_SWITCH
     __cilk_tbb_unwatch_thunk &ut = s->my_cilk_unwatch_thunk;
     if ( ut.routine )
@@ -148,43 +155,94 @@ void governor::sign_off(generic_scheduler* s) {
 #endif /* __TBB_SURVIVE_THREAD_SWITCH */
 }
 
-generic_scheduler* governor::init_scheduler( unsigned num_threads, stack_size_type stack_size, bool auto_init ) {
+void governor::setBlockingTerminate(const task_scheduler_init *tsi) {
+    __TBB_ASSERT(!IsBlockingTerminationInProgress, "It's impossible to create task_scheduler_init while blocking termination is in progress.");
+    if (BlockingTSI)
+        throw_exception(eid_blocking_sch_init);
+    BlockingTSI = tsi;
+}
+
+void governor::one_time_init() {
     if( !__TBB_InitOnce::initialization_done() )
         DoOneTimeInitializations();
-    generic_scheduler* s = theTLS.get();
-    if( s ) {
-        s->my_ref_count += 1;
-        return s;
-    }
 #if __TBB_SURVIVE_THREAD_SWITCH
     atomic_do_once( &initialize_cilk_interop, cilkrts_load_state );
 #endif /* __TBB_SURVIVE_THREAD_SWITCH */
-    if( (int)num_threads == task_scheduler_init::automatic )
+}
+
+generic_scheduler* governor::init_scheduler_weak() {
+    one_time_init();
+    __TBB_ASSERT( is_set(NULL), "TLS contains a scheduler?" );
+    generic_scheduler* s = generic_scheduler::create_master( NULL ); // without arena
+    s->my_auto_initialized = true;
+    return s;
+}
+
+generic_scheduler* governor::init_scheduler( int num_threads, stack_size_type stack_size, bool auto_init ) {
+    one_time_init();
+    uintptr_t v = theTLS.get();
+    generic_scheduler* s = tls_scheduler_of( v );
+    if( v&1 ) { // TLS holds scheduler instance with arena
+        __TBB_ASSERT( s->my_arena, "TLS is marked for scheduler with arena" );
+        s->my_ref_count += 1;
+        return s;
+    }
+    if( v ) { //TLS holds scheduler instance without arena
+        __TBB_ASSERT( !s->my_arena, "TLS is marked for scheduler without arena" );
+        __TBB_ASSERT( s->my_auto_initialized, "weakly initialized scheduler is supposed to be auto-initialized" );
+        s->attach_arena( market::create_arena( default_num_threads(), 0, true ), 0, /*is_master*/true );
+        __TBB_ASSERT( s->my_arena_index == 0, "Master thread must occupy the first slot in its arena" );
+        s->my_arena_slot->my_scheduler = s;
+        s->my_arena->my_default_ctx = s->default_context(); // it also transfers implied ownership
+        s->my_ref_count += 1;
+        assume_scheduler( s );
+        return s;
+    }
+    // Create new scheduler instance with arena
+    bool default_concurrency_requested = num_threads == task_scheduler_init::automatic;
+    if( default_concurrency_requested )
         num_threads = default_num_threads();
-    s = generic_scheduler::create_master( 
-            market::create_arena( num_threads - 1, stack_size ? stack_size : ThreadStackSize ) );
+    arena *a = market::create_arena( num_threads, stack_size, default_concurrency_requested );
+    s = generic_scheduler::create_master( a );
     __TBB_ASSERT(s, "Somehow a local scheduler creation for a master thread failed");
+    __TBB_ASSERT( is_set(s), NULL );
     s->my_auto_initialized = auto_init;
     return s;
 }
 
-void governor::terminate_scheduler( generic_scheduler* s ) {
-    __TBB_ASSERT( s == theTLS.get(), "Attempt to terminate non-local scheduler instance" );
-    if( !--(s->my_ref_count) )
+void governor::terminate_scheduler( generic_scheduler* s, const task_scheduler_init* tsi_ptr ) {
+    __TBB_ASSERT( is_set(s), "Attempt to terminate non-local scheduler instance" );
+    if (--(s->my_ref_count)) {
+        // can't throw exception, because this is on dtor's call chain
+        __TBB_ASSERT_RELEASE( !BlockingTSI || BlockingTSI!=tsi_ptr,
+                              "Attempt to terminate nested scheduler in blocking mode" );
+    } else {
+#if TBB_USE_ASSERT
+        if (BlockingTSI) {
+            __TBB_ASSERT( BlockingTSI == tsi_ptr, "For blocking termination last terminate_scheduler must be blocking." );
+            IsBlockingTerminationInProgress = true;
+        }
+#endif
         s->cleanup_master();
+        __TBB_ASSERT( is_set(NULL), "cleanup_master has not cleared its TLS slot" );
+        BlockingTSI = NULL;
+#if TBB_USE_ASSERT
+        IsBlockingTerminationInProgress = false;
+#endif
+    }
 }
 
 void governor::auto_terminate(void* arg){
-    generic_scheduler* s = static_cast<generic_scheduler*>(arg);
+    generic_scheduler* s = tls_scheduler_of( uintptr_t(arg) ); // arg is equivalent to theTLS.get()
     if( s && s->my_auto_initialized ) {
         if( !--(s->my_ref_count) ) {
+            __TBB_ASSERT( !BlockingTSI, "Blocking auto-terminate is not supported." );
             // If the TLS slot is already cleared by OS or underlying concurrency
             // runtime, restore its value.
-            if ( !theTLS.get() )
-                theTLS.set(s);
-            else __TBB_ASSERT( s == theTLS.get(), NULL );
+            if( !is_set(s) )
+                assume_scheduler(s);
             s->cleanup_master();
-            __TBB_ASSERT( !theTLS.get(), "cleanup_master has not cleared its TLS slot" );
+            __TBB_ASSERT( is_set(NULL), "cleanup_master has not cleared its TLS slot" );
         }
     }
 }
@@ -212,7 +270,7 @@ __cilk_tbb_retcode governor::stack_op_handler( __cilk_tbb_stack_op op, void* dat
     __TBB_ASSERT(data,NULL);
     generic_scheduler* s = static_cast<generic_scheduler*>(data);
 #if TBB_USE_ASSERT
-    void* current = theTLS.get();
+    void* current = local_scheduler_if_initialized();
 #if _WIN32||_WIN64
     uintptr_t thread_id = GetCurrentThreadId();
 #else
@@ -231,7 +289,7 @@ __cilk_tbb_retcode governor::stack_op_handler( __cilk_tbb_stack_op op, void* dat
                 runtime_warning( "redundant adoption of %p by thread %p\n", s, (void*)thread_id );
             s->my_cilk_state = generic_scheduler::cs_running;
 #endif /* TBB_USE_ASSERT */
-            theTLS.set(s);
+            assume_scheduler( s );
             break;
         }
         case CILK_TBB_STACK_ORPHAN: {
@@ -239,7 +297,7 @@ __cilk_tbb_retcode governor::stack_op_handler( __cilk_tbb_stack_op op, void* dat
 #if TBB_USE_ASSERT
             s->my_cilk_state = generic_scheduler::cs_limbo;
 #endif /* TBB_USE_ASSERT */
-            theTLS.set(NULL);
+            assume_scheduler(NULL);
             break;
         }
         case CILK_TBB_STACK_RELEASE: {
@@ -275,10 +333,17 @@ void task_scheduler_init::initialize( int number_of_threads, stack_size_type thr
 #endif
     thread_stack_size &= ~(stack_size_type)propagation_mode_mask;
     if( number_of_threads!=deferred ) {
-        __TBB_ASSERT( !my_scheduler, "task_scheduler_init already initialized" );
-        __TBB_ASSERT( number_of_threads==-1 || number_of_threads>=1,
-                    "number_of_threads for task_scheduler_init must be -1 or positive" );
-        internal::generic_scheduler *s = governor::init_scheduler( number_of_threads, thread_stack_size );
+        bool blocking_terminate = false;
+        if (my_scheduler == (scheduler*)wait_workers_in_terminate_flag) {
+            blocking_terminate = true;
+            my_scheduler = NULL;
+        }
+        __TBB_ASSERT_RELEASE( !my_scheduler, "task_scheduler_init already initialized" );
+        __TBB_ASSERT_RELEASE( number_of_threads==automatic || number_of_threads > 0,
+                    "number_of_threads for task_scheduler_init must be automatic or positive" );
+        if (blocking_terminate)
+            governor::setBlockingTerminate(this);
+        internal::generic_scheduler *s = governor::init_scheduler( number_of_threads, thread_stack_size, /*auto_init=*/false );
 #if __TBB_TASK_GROUP_CONTEXT && TBB_USE_EXCEPTIONS
         if ( s->master_outermost_level() ) {
             uintptr_t &vt = s->default_context()->my_version_and_traits;
@@ -294,7 +359,7 @@ void task_scheduler_init::initialize( int number_of_threads, stack_size_type thr
 #endif /* __TBB_TASK_GROUP_CONTEXT && TBB_USE_EXCEPTIONS */
             my_scheduler = s;
     } else {
-        __TBB_ASSERT( !thread_stack_size, "deferred initialization ignores stack size setting" );
+        __TBB_ASSERT_RELEASE( !thread_stack_size, "deferred initialization ignores stack size setting" );
     }
 }
 
@@ -305,7 +370,7 @@ void task_scheduler_init::terminate() {
 #endif /* __TBB_TASK_GROUP_CONTEXT && TBB_USE_EXCEPTIONS */
     generic_scheduler* s = static_cast<generic_scheduler*>(my_scheduler);
     my_scheduler = NULL;
-    __TBB_ASSERT( s, "task_scheduler_init::terminate without corresponding task_scheduler_init::initialize()");
+    __TBB_ASSERT_RELEASE( s, "task_scheduler_init::terminate without corresponding task_scheduler_init::initialize()");
 #if __TBB_TASK_GROUP_CONTEXT && TBB_USE_EXCEPTIONS
     if ( s->master_outermost_level() ) {
         uintptr_t &vt = s->default_context()->my_version_and_traits;
@@ -313,7 +378,7 @@ void task_scheduler_init::terminate() {
                                         : vt & ~task_group_context::exact_exception;
     }
 #endif /* __TBB_TASK_GROUP_CONTEXT && TBB_USE_EXCEPTIONS */
-    governor::terminate_scheduler(s);
+    governor::terminate_scheduler(s, this);
 }
 
 int task_scheduler_init::default_num_threads() {
