@@ -1,29 +1,21 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
 
-    This file is part of Threading Building Blocks.
+    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
+    you can redistribute it and/or modify it under the terms of the GNU General Public License
+    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
+    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
+    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+    See  the GNU General Public License for more details.   You should have received a copy of
+    the  GNU General Public License along with Threading Building Blocks; if not, write to the
+    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
 
-    Threading Building Blocks is free software; you can redistribute it
-    and/or modify it under the terms of the GNU General Public License
-    version 2 as published by the Free Software Foundation.
-
-    Threading Building Blocks is distributed in the hope that it will be
-    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with Threading Building Blocks; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-
-    As a special exception, you may use this file as part of a free software
-    library without restriction.  Specifically, if other files instantiate
-    templates or use macros or inline functions from this file, or you compile
-    this file and link it with other files to produce an executable, this
-    file does not by itself cause the resulting executable to be covered by
-    the GNU General Public License.  This exception does not however
-    invalidate any other reasons why the executable file might be covered by
-    the GNU General Public License.
+    As a special exception,  you may use this file  as part of a free software library without
+    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
+    functions from this file, or you compile this file and link it with other files to produce
+    an executable,  this file does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however invalidate any other
+    reasons why the executable file might be covered by the GNU General Public License.
 */
 
 #ifndef _TBB_market_H
@@ -33,7 +25,7 @@
 
 #include "scheduler_common.h"
 #include "tbb/atomic.h"
-#include "tbb/spin_mutex.h"
+#include "tbb/spin_rw_mutex.h"
 #include "../rml/include/rml_tbb.h"
 
 #include "intrusive_list.h"
@@ -50,10 +42,6 @@ class task_group_context;
 
 namespace internal {
 
-class arena;
-class generic_scheduler;
-template<typename SchedulerTraits> class custom_scheduler;
-
 //------------------------------------------------------------------------
 // Class market
 //------------------------------------------------------------------------
@@ -61,12 +49,14 @@ template<typename SchedulerTraits> class custom_scheduler;
 class market : no_copy, rml::tbb_client {
     friend class generic_scheduler;
     friend class arena;
+    friend class tbb::interface7::internal::task_arena_base;
     template<typename SchedulerTraits> friend class custom_scheduler;
     friend class tbb::task_group_context;
 private:
     friend void ITT_DoUnsafeOneTimeInitialization ();
 
     typedef intrusive_list<arena> arena_list_type;
+    typedef intrusive_list<generic_scheduler> scheduler_list_type;
 
     //! Currently active global market
     static market* theMarket;
@@ -76,26 +66,31 @@ private:
     //! Mutex guarding creation/destruction of theMarket, insertions/deletions in my_arenas, and cancellation propagation
     static global_market_mutex_type  theMarketMutex;
 
-    //! Reference count controlling market object lifetime
-    intptr_t my_ref_count;
-
     //! Lightweight mutex guarding accounting operations with arenas list
-    typedef scheduler_mutex_type arenas_list_mutex_type;
+    typedef spin_rw_mutex arenas_list_mutex_type;
     arenas_list_mutex_type my_arenas_list_mutex;
 
     //! Pointer to the RML server object that services this TBB instance.
     rml::tbb_server* my_server;
 
-    //! Stack size of worker threads
-    size_t my_stack_size;
+    //! Maximal number of workers allowed for use by the underlying resource manager
+    /** It can't be changed after market creation. **/
+    unsigned my_num_workers_hard_limit;
 
-    //! Number of workers requested from the underlying resource manager
-    unsigned my_max_num_workers;
+    //! Current application-imposed limit on the number of workers (see set_active_num_workers())
+    /** It can't be more than my_num_workers_hard_limit. **/
+    unsigned my_num_workers_soft_limit;
 
-    //! Number of workers that have been delivered by RML
+    //! Number of workers currently requested from RML
+    int my_num_workers_requested;
+
+    //! First unused index of worker
     /** Used to assign indices to the new workers coming from RML, and busy part
         of my_workers array. **/
-    atomic<unsigned> my_num_workers;
+    atomic<unsigned> my_first_unused_worker_idx;
+
+    //! Number of workers that were requested by all arenas
+    int my_total_demand;
 
 #if __TBB_TASK_PRIORITY
     //! Highest priority among active arenas in the market.
@@ -121,28 +116,18 @@ private:
 
         //! The first arena to be checked when idle worker seeks for an arena to enter
         /** The check happens in round-robin fashion. **/
-        arena_list_type::iterator next_arena;
+        arena *next_arena;
 
         //! Total amount of workers requested by arenas at this priority level.
         int workers_requested;
 
         //! Maximal amount of workers the market can tell off to this priority level.
         int workers_available;
-
-#if __TBB_TRACK_PRIORITY_LEVEL_SATURATION
-        //! Total amount of workers that are in arenas at this priority level.
-        int workers_present;
-#endif /* __TBB_TRACK_PRIORITY_LEVEL_SATURATION */
     }; // struct priority_level_info
 
     //! Information about arenas at different priority levels
     priority_level_info my_priority_levels[num_priority_levels];
 
-#if __TBB_TRACK_PRIORITY_LEVEL_SATURATION
-    //! Lowest priority level having workers available.
-    intptr_t my_lowest_populated_level;
-#endif /* __TBB_TRACK_PRIORITY_LEVEL_SATURATION */
-
 #else /* !__TBB_TASK_PRIORITY */
 
     //! List of registered arenas
@@ -150,15 +135,24 @@ private:
 
     //! The first arena to be checked when idle worker seeks for an arena to enter
     /** The check happens in round-robin fashion. **/
-    arena_list_type::iterator my_next_arena;
-
-    //! Number of workers that were requested by all arenas
-    int my_total_demand;
+    arena *my_next_arena;
 #endif /* !__TBB_TASK_PRIORITY */
 
     //! ABA prevention marker to assign to newly created arenas
     uintptr_t my_arenas_aba_epoch;
 
+    //! Reference count controlling market object lifetime
+    unsigned my_ref_count;
+
+    //! Count of master threads attached
+    unsigned my_public_ref_count;
+
+    //! Stack size of worker threads
+    size_t my_stack_size;
+
+    //! Shutdown mode
+    bool join_workers;
+
 #if __TBB_COUNT_TASK_NODES
     //! Net number of nodes that have been allocated from heap.
     /** Updated each time a scheduler or arena is destroyed. */
@@ -166,23 +160,18 @@ private:
 #endif /* __TBB_COUNT_TASK_NODES */
 
     //! Constructor
-    market ( unsigned max_num_workers, size_t stack_size );
+    market ( unsigned workers_soft_limit, unsigned workers_hard_limit, size_t stack_size );
 
     //! Factory method creating new market object
-    static market& global_market ( unsigned max_num_workers, size_t stack_size );
+    static market& global_market ( unsigned max_num_workers = 0, size_t stack_size = 0,
+                                   bool default_concurrency_requested = false, bool is_public = false );
 
     //! Destroys and deallocates market object created by market::create()
     void destroy ();
 
-    void try_destroy_arena ( arena*, uintptr_t aba_epoch );
-
 #if __TBB_TASK_PRIORITY
     //! Returns next arena that needs more workers, or NULL.
-    arena* arena_in_need (
-#if __TBB_TRACK_PRIORITY_LEVEL_SATURATION
-                           arena* prev_arena
-#endif /* __TBB_TRACK_PRIORITY_LEVEL_SATURATION */
-                         );
+    arena* arena_in_need ( arena* prev_arena );
 
     //! Recalculates the number of workers assigned to each arena at and below the specified priority.
     /** The actual number of workers servicing a particular arena may temporarily 
@@ -220,8 +209,10 @@ private:
     }
 
     //! Returns next arena that needs more workers, or NULL.
-    arena* arena_in_need () {
-        spin_mutex::scoped_lock lock(my_arenas_list_mutex);
+    arena* arena_in_need (arena*) {
+        if(__TBB_load_with_acquire(my_total_demand) <= 0)
+            return NULL;
+        arenas_list_mutex_type::scoped_lock lock(my_arenas_list_mutex, /*is_writer=*/false);
         return arena_in_need(my_arenas, my_next_arena);
     }
     void assert_market_valid () const {}
@@ -230,7 +221,6 @@ private:
     //! Returns number of masters doing computational (CPU-intensive) work
     int num_active_masters () { return 1; }  // APM TODO: replace with a real mechanism
 
-
     ////////////////////////////////////////////////////////////////////////////////
     // Helpers to unify code branches dependent on priority feature presence
 
@@ -238,7 +228,7 @@ private:
 
     void remove_arena_from_list ( arena& a );
 
-    arena* arena_in_need ( arena_list_type &arenas, arena_list_type::iterator& next );
+    arena* arena_in_need ( arena_list_type &arenas, arena *&next );
 
     static void update_allotment ( arena_list_type& arenas, int total_demand, int max_workers );
 
@@ -248,7 +238,7 @@ private:
 
     /*override*/ version_type version () const { return 0; }
 
-    /*override*/ unsigned max_job_count () const { return my_max_num_workers; }
+    /*override*/ unsigned max_job_count () const { return my_num_workers_hard_limit; }
 
     /*override*/ size_t min_stack_size () const { return worker_stack_size(); }
 
@@ -266,31 +256,35 @@ public:
     //! Creates an arena object
     /** If necessary, also creates global market instance, and boosts its ref count.
         Each call to create_arena() must be matched by the call to arena::free_arena(). **/
-    static arena& create_arena ( unsigned max_num_workers, size_t stack_size );
+    static arena* create_arena ( int num_slots, size_t stack_size, bool default_concurrency_requested );
 
     //! Removes the arena from the market's list
-    static void try_destroy_arena ( market*, arena*, uintptr_t aba_epoch, bool master );
+    void try_destroy_arena ( arena*, uintptr_t aba_epoch );
 
     //! Removes the arena from the market's list
     void detach_arena ( arena& );
 
     //! Decrements market's refcount and destroys it in the end
-    void release ();
+    void release ( bool is_public = false );
 
     //! Request that arena's need in workers should be adjusted.
     /** Concurrent invocations are possible only on behalf of different arenas. **/
     void adjust_demand ( arena&, int delta );
 
-    //! Guarantee that request_close_connection() is called by master, not some worker
-    /** Must be called before arena::on_thread_leaving() **/
-    void prepare_wait_workers() { ++my_ref_count; }
-
-    //! Wait workers termiantion
+    //! Wait workers termination
     void wait_workers ();
 
+    bool must_join_workers () const { return join_workers; }
+
     //! Returns the requested stack size of worker threads.
     size_t worker_stack_size () const { return my_stack_size; }
 
+    //! Set number of active workers
+    static void set_active_num_workers( unsigned w );
+
+    //! Reports active parallelism level according to user's settings
+    static unsigned app_parallelism_limit();
+
 #if _WIN32||_WIN64
     //! register master with the resource manager
     void register_master( ::rml::server::execution_resource_t& rsc_handle ) {
@@ -307,6 +301,10 @@ public:
 
 #if __TBB_TASK_GROUP_CONTEXT
     //! Finds all contexts affected by the state change and propagates the new state to them.
+    /** The propagation is relayed to the market because tasks created by one
+        master thread can be passed to and executed by other masters. This means
+        that context trees can span several arenas at once and thus state change
+        propagation cannot be generally localized to one arena only. **/
     template <typename T>
     bool propagate_task_group_state ( T task_group_context::*mptr_state, task_group_context& src, T new_state );
 #endif /* __TBB_TASK_GROUP_CONTEXT */
@@ -327,43 +325,27 @@ public:
 #endif /* __TBB_TASK_PRIORITY */
 
 #if __TBB_COUNT_TASK_NODES
-    //! Returns the number of task objects "living" in worker threads
-    intptr_t workers_task_node_count();
-
     //! Net number of nodes that have been allocated from heap.
     /** Updated each time a scheduler or arena is destroyed. */
     void update_task_node_count( intptr_t delta ) { my_task_node_count += delta; }
 #endif /* __TBB_COUNT_TASK_NODES */
 
 #if __TBB_TASK_GROUP_CONTEXT
+    //! List of registered master threads
+    scheduler_list_type my_masters;
+
     //! Array of pointers to the registered workers
     /** Used by cancellation propagation mechanism.
         Must be the last data member of the class market. **/
     generic_scheduler* my_workers[1];
 #endif /* __TBB_TASK_GROUP_CONTEXT */
 
+    static unsigned max_num_workers() {
+        global_market_mutex_type::scoped_lock lock( theMarketMutex );
+        return theMarket? theMarket->my_num_workers_hard_limit : 0;
+    }
 }; // class market
 
-#if __TBB_TASK_PRIORITY
-    #define BeginForEachArena(a)    \
-        arenas_list_mutex_type::scoped_lock arena_list_lock(my_arenas_list_mutex);  \
-        for ( intptr_t i = my_global_top_priority; i >= my_global_bottom_priority; --i ) {  \
-            /*arenas_list_mutex_type::scoped_lock arena_list_lock(my_priority_levels[i].my_arenas_list_mutex);*/ \
-            arena_list_type &arenas = my_priority_levels[i].arenas;
-#else /* !__TBB_TASK_PRIORITY */
-    #define BeginForEachArena(a)    \
-        arena_list_type &arenas = my_arenas; {
-#endif /* !__TBB_TASK_PRIORITY */
-
-#define ForEachArena(a)     \
-    BeginForEachArena(a)    \
-        arena_list_type::iterator it = arenas.begin();  \
-        for ( ; it != arenas.end(); ++it ) {            \
-            arena &a = *it;
-
-#define EndForEach() }}
-
-
 } // namespace internal
 } // namespace tbb
 
