@@ -35,6 +35,9 @@
     #error Must define USE_PTHREAD or USE_WINTHREAD
 #endif
 
+// TODO: *BSD also has it
+#define BACKEND_HAS_MREMAP __linux__
+
 #include "tbb/tbb_config.h" // for __TBB_LIBSTDCPP_EXCEPTION_HEADERS_BROKEN
 #if __TBB_LIBSTDCPP_EXCEPTION_HEADERS_BROKEN
   #define _EXCEPTION_PTR_H /* prevents exception_ptr.h inclusion */
@@ -109,19 +112,6 @@ const uintptr_t slabSize = 16*1024;
  */
 const unsigned cacheCleanupFreq = 256;
 
-/*
- * Best estimate of cache line size, for the purpose of avoiding false sharing.
- * Too high causes memory overhead, too low causes false-sharing overhead.
- * Because, e.g., 32-bit code might run on a 64-bit system with a larger cache line size,
- * it would probably be better to probe at runtime where possible and/or allow for an environment variable override,
- * but currently this is still used for compile-time layout of class Block, so the change is not entirely trivial.
- */
-#if __powerpc64__ || __ppc64__ || __bgp__
-const uint32_t estimatedCacheLineSize = 128;
-#else
-const uint32_t estimatedCacheLineSize =  64;
-#endif
-
 /*
  * Alignment of large (>= minLargeObjectSize) objects.
  */
@@ -143,8 +133,8 @@ extern const uint32_t minLargeObjectSize;
 class TLSKey {
     tls_key_t TLS_pointer_key;
 public:
-    TLSKey();
-   ~TLSKey();
+    bool init();
+    bool destroy();
     TLSData* getThreadMallocTLS() const;
     void setThreadMallocTLS( TLSData * newvalue );
     TLSData* createTLS(MemoryPool *memPool, Backend *backend);
@@ -459,9 +449,7 @@ public:
             bin[i].init();
         bitMask.reset();
     }
-#if __TBB_MALLOC_LOCACHE_STAT
     void reportStat(FILE *f);
-#endif
 #if __TBB_MALLOC_WHITEBOX_TEST
     size_t getLOCSize() const;
     size_t getUsedSize() const;
@@ -521,9 +509,7 @@ public:
         largeCache.reset();
         hugeCache.reset();
     }
-#if __TBB_MALLOC_LOCACHE_STAT
     void reportStat(FILE *f);
-#endif
 #if __TBB_MALLOC_WHITEBOX_TEST
     size_t getLOCSize() const;
     size_t getUsedSize() const;
@@ -560,6 +546,7 @@ class BlockI {
 };
 
 struct LargeMemoryBlock : public BlockI {
+    MemoryPool       *pool;          // owner pool
     LargeMemoryBlock *next,          // ptrs in list of cached blocks
                      *prev,
     // 2-linked list of pool's large objects
@@ -569,7 +556,7 @@ struct LargeMemoryBlock : public BlockI {
                      *gNext;
     uintptr_t         age;           // age of block while in cache
     size_t            objectSize;    // the size requested by a client
-    size_t            unalignedSize; // the size requested from getMemory
+    size_t            unalignedSize; // the size requested from backend
     BackRefIdx        backRefIdx;    // cached here, used copy is in LargeObjectHdr
 };
 
@@ -577,52 +564,40 @@ struct LargeMemoryBlock : public BlockI {
 class BackendSync {
     // Class instances should reside in zero-initialized memory!
     // The number of blocks currently removed from a bin and not returned back
-    intptr_t  blocksInProcessing;  // to another
-    intptr_t  binsModifications;   // incremented on every bin modification
+    intptr_t inFlyBlocks;         // to another
+    intptr_t binsModifications;   // incremented on every bin modification
+    Backend *backend;
 public:
-    void blockConsumed() { AtomicIncrement(blocksInProcessing); }
+    void init(Backend *b) { backend = b; }
+    void blockConsumed() { AtomicIncrement(inFlyBlocks); }
     void binsModified() { AtomicIncrement(binsModifications); }
     void blockReleased() {
 #if __TBB_MALLOC_BACKEND_STAT
-        MALLOC_ITT_SYNC_RELEASING(&blocksInProcessing);
+        MALLOC_ITT_SYNC_RELEASING(&inFlyBlocks);
 #endif
         AtomicIncrement(binsModifications);
-        intptr_t prev = AtomicAdd(blocksInProcessing, -1);
+        intptr_t prev = AtomicAdd(inFlyBlocks, -1);
         MALLOC_ASSERT(prev > 0, ASSERT_TEXT);
         suppress_unused_warning(prev);
     }
     intptr_t getNumOfMods() const { return FencedLoad(binsModifications); }
     // return true if need re-do the blocks search
-    bool waitTillBlockReleased(intptr_t startModifiedCnt) {
-#if __TBB_MALLOC_BACKEND_STAT
-        MALLOC_ITT_SYNC_PREPARE(&blocksInProcessing);
-#endif
-        for (intptr_t myBlocksNum = FencedLoad(blocksInProcessing);
-             // no blocks in processing, stop waiting
-             myBlocksNum; ) {
-            SpinWaitWhileEq(blocksInProcessing, myBlocksNum);
-            WhiteboxTestingYield();
-            intptr_t newBlocksNum = FencedLoad(blocksInProcessing);
-            // stop waiting iff blocks were removed from processing,
-            // if blocks were added, there is no reason to stop waiting
-            if (newBlocksNum < myBlocksNum)
-                break;
-            myBlocksNum = newBlocksNum;
-        }
-#if __TBB_MALLOC_BACKEND_STAT
-        MALLOC_ITT_SYNC_ACQUIRED(&blocksInProcessing);
-#endif
-        // were bins modified since scanned?
-        return startModifiedCnt != getNumOfMods();
-    }
+    inline bool waitTillBlockReleased(intptr_t startModifiedCnt);
 };
 
 class CoalRequestQ { // queue of free blocks that coalescing was delayed
 private:
-    FreeBlock *blocksToFree;
+    FreeBlock   *blocksToFree;
+    BackendSync *bkndSync;
+    // counted blocks in blocksToFree and that are leaved blocksToFree
+    // and still in active coalescing
+    intptr_t     inFlyBlocks;
 public:
+    void init(BackendSync *bSync) { bkndSync = bSync; }
     FreeBlock *getAll(); // return current list of blocks and make queue empty
     void putBlock(FreeBlock *fBlock);
+    inline void blockWasProcessed();
+    intptr_t blocksInFly() const { return FencedLoad(inFlyBlocks); }
 };
 
 class MemExtendingSema {
@@ -659,6 +634,15 @@ enum MemRegionType {
     MEMREG_ONE_BLOCK
 };
 
+class MemRegionList {
+    MallocMutex regionListLock;
+public:
+    MemRegion  *head;
+    void add(MemRegion *r);
+    void remove(MemRegion *r);
+    int reportStat(FILE *f);
+};
+
 class Backend {
 private:
 /* Blocks in range [minBinnedSize; getMaxBinnedSize()] are kept in bins,
@@ -699,11 +683,11 @@ public:
 
         void removeBlock(FreeBlock *fBlock);
         void reset() { head = tail = 0; }
-#if __TBB_MALLOC_BACKEND_STAT
-        size_t countFreeBlocks();
-        void reportStat(FILE *f);
-#endif
         bool empty() const { return !head; }
+
+        size_t countFreeBlocks();
+        size_t reportFreeBlocks(FILE *f);
+        void reportStat(FILE *f);
     };
 
     typedef BitMaskMin<Backend::freeBinsNum> BitMaskBins;
@@ -727,10 +711,8 @@ public:
             return p == -1 ? Backend::freeBinsNum : p;
         }
         void verify();
-#if __TBB_MALLOC_BACKEND_STAT
-        void reportStat(FILE *f);
-#endif
         void reset();
+        void reportStat(FILE *f);
     };
 
 private:
@@ -751,8 +733,7 @@ private:
 
     ExtMemoryPool *extMemPool;
     // used for release every region on pool destroying
-    MemRegion     *regionList;
-    MallocMutex    regionListLock;
+    MemRegionList  regionList;
 
     CoalRequestQ   coalescQ; // queue of coalescing requests
     BackendSync    bkndSync;
@@ -760,11 +741,14 @@ private:
     MemExtendingSema memExtendingSema;
     size_t         totalMemSize,
                    memSoftLimit;
-    // Fixed pools request memory once per lifetime, during pool_create.
-    // Status of memory acquisition for such pool keeps here.
-    // So value is changed only for fixed pools, and without synchronization,
-    // as pool is not available till returning from pool_create.
-    bool           rawMemReceived;
+    // to keep 1st allocation large than requested, keep bootstrapping status
+    enum {
+        bootsrapMemNotDone = 0,
+        bootsrapMemInitializing,
+        bootsrapMemDone
+    };
+    intptr_t       bootsrapMemStatus;
+    MallocMutex    bootsrapMemStatusMutex;
 
     // Using of maximal observed requested size allows decrease
     // memory consumption for small requests and decrease fragmentation
@@ -777,6 +761,9 @@ private:
     void startUseBlock(MemRegion *region, FreeBlock *fBlock, bool addToBin);
     void releaseRegion(MemRegion *region);
 
+    FreeBlock *releaseMemInCaches(intptr_t startModifiedCnt,
+                                  int *lockedBinsThreshold, int numOfLockedBins);
+    void requestBootstrapMem();
     FreeBlock *askMemFromOS(size_t totalReqSize, intptr_t startModifiedCnt,
                             int *lockedBinsThreshold, int numOfLockedBins,
                             bool *splittable);
@@ -788,26 +775,25 @@ private:
                             bool needAlignedRes);
 
     FreeBlock *doCoalesc(FreeBlock *fBlock, MemRegion **memRegion);
-    bool coalescAndPutList(FreeBlock *head, bool forceCoalescQDrop);
-    bool scanCoalescQ(bool forceCoalescQDrop);
+    bool coalescAndPutList(FreeBlock *head, bool forceCoalescQDrop, bool reportBlocksProcessed);
     void coalescAndPut(FreeBlock *fBlock, size_t blockSz);
 
     void removeBlockFromBin(FreeBlock *fBlock);
 
     void *allocRawMem(size_t &size) const;
-    void freeRawMem(void *object, size_t size) const;
+    bool freeRawMem(void *object, size_t size) const;
 
     void putLargeBlock(LargeMemoryBlock *lmb);
     void releaseCachesToLimit();
 public:
+    bool scanCoalescQ(bool forceCoalescQDrop);
+    intptr_t blocksInCoalescing() const { return coalescQ.blocksInFly(); }
     void verify();
-#if __TBB_MALLOC_BACKEND_STAT
-    void reportStat(FILE *f);
-#endif
-    bool bootstrap(ExtMemoryPool *extMemoryPool);
+    void init(ExtMemoryPool *extMemoryPool);
     void reset();
     bool destroy();
     bool clean(); // clean on caches cleanup
+    void reportStat(FILE *f);
 
     BlockI *getSlabBlock(int num) {
         BlockI *b = (BlockI*)
@@ -826,6 +812,8 @@ public:
     LargeMemoryBlock *getLargeBlock(size_t size);
     void returnLargeObject(LargeMemoryBlock *lmb);
 
+    void *remap(void *ptr, size_t oldSize, size_t newSize, size_t alignment);
+
     void setRecommendedMaxSize(size_t softLimit) {
         memSoftLimit = softLimit;
         releaseCachesToLimit();
@@ -895,7 +883,7 @@ struct ExtMemoryPool {
 
     bool init(intptr_t poolId, rawAllocType rawAlloc, rawFreeType rawFree,
               size_t granularity, bool keepAllMemory, bool fixedPool);
-    void initTLS();
+    bool initTLS();
 
     // i.e., not system default pool for scalable_malloc/scalable_free
     bool userPool() const { return rawAlloc; }
@@ -904,29 +892,37 @@ struct ExtMemoryPool {
     bool softCachesCleanup();
     bool releaseAllLocalCaches();
     bool hardCachesCleanup();
-    void reset() {
+    bool reset() {
         loc.reset();
         allLocalCaches.reset();
-        tlsPointerKey.~TLSKey();
+        bool ret = tlsPointerKey.destroy();
         backend.reset();
+        return ret;
     }
-    void destroy() {
+    bool destroy() {
+        MALLOC_ASSERT(isPoolValid(),
+                      "Possible double pool_destroy or heap corruption");
         if (!userPool()) {
             loc.reset();
             allLocalCaches.reset();
         }
         // pthread_key_dtors must be disabled before memory unmapping
         // TODO: race-free solution
-        tlsPointerKey.~TLSKey();
+        bool ret = tlsPointerKey.destroy();
         if (rawFree || !userPool())
-            backend.destroy();
+            ret &= backend.destroy();
+        // pool is not valid after this point
+        granularity = 0;
+        return ret;
     }
     void delayRegionsReleasing(bool mode) { delayRegsReleasing = mode; }
     inline bool regionsAreReleaseable() const;
 
-    LargeMemoryBlock *mallocLargeObject(size_t allocationSize);
+    LargeMemoryBlock *mallocLargeObject(MemoryPool *pool, size_t allocationSize);
     void freeLargeObject(LargeMemoryBlock *lmb);
     void freeLargeObjectList(LargeMemoryBlock *head);
+    // use granulatity as marker for pool validity
+    bool isPoolValid() const { return granularity; }
 };
 
 inline bool Backend::inUserPool() const { return extMemPool->userPool(); }
@@ -990,7 +986,7 @@ public:
     }
     void printStatus();
     void registerAllocation(bool available);
-    void registerReleasing(size_t size);
+    void registerReleasing(void* addr, size_t size);
 
     void init(size_t hugePageSize) {
         MALLOC_ASSERT(!hugePageSize || isPowerOfTwo(hugePageSize),
