@@ -83,7 +83,11 @@ int TestMain ();
 #if _WIN32||_WIN64
     #include "tbb/machine/windows_api.h"
     #if _WIN32_WINNT > 0x0501 && _MSC_VER && !_M_ARM
+        // Suppress "typedef ignored ... when no variable is declared" warning by vc14
+        #pragma warning (push)
+        #pragma warning (disable: 4091)
         #include <dbghelp.h>
+        #pragma warning (pop)
         #pragma comment (lib, "dbghelp.lib")
     #endif
     #if _XBOX
@@ -144,7 +148,7 @@ void print_call_stack() {
             if(!SymFromAddr( GetCurrentProcess(), DWORD64(buff[i]), &offset, &sym )) {
                 sym.Address = ULONG64(buff[i]); offset = 0; sym.Name[0] = 0;
             }
-            REPORT("[%d] %016I64LX+%04I64LX: %s\n", i, sym.Address, offset, sym.Name); //TODO: print module name
+            REPORT("[%d] %016I64X+%04I64X: %s\n", i, sym.Address, offset, sym.Name); //TODO: print module name
         }
     #endif /*BACKTRACE_FUNCTION_AVAILABLE*/
 }
