@@ -289,7 +289,7 @@ struct harness_counting_receiver : public tbb::flow::receiver<T>, NoCopy {
 
     /* override */ tbb::task *try_put_task( const T & ) {
       ++my_count;
-      return const_cast<tbb::task *>(tbb::flow::interface7::SUCCESSFULLY_ENQUEUED);
+      return const_cast<tbb::task *>(tbb::flow::interface8::SUCCESSFULLY_ENQUEUED);
     }
 
     void validate() {
@@ -298,16 +298,17 @@ struct harness_counting_receiver : public tbb::flow::receiver<T>, NoCopy {
     }
 
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+    typedef typename tbb::flow::receiver<T>::built_predecessors_type built_predecessors_type;
+    built_predecessors_type mbp;
+    /*override*/ built_predecessors_type &built_predecessors() { return mbp; }
     typedef typename tbb::flow::receiver<T>::predecessor_list_type predecessor_list_type;
     /*override*/void internal_add_built_predecessor(tbb::flow::sender<T> &) {}
     /*override*/void internal_delete_built_predecessor(tbb::flow::sender<T> &) {}
     /*override*/void copy_predecessors(predecessor_list_type &) { }
     /*override*/size_t predecessor_count() { return 0; }
-    /*override*/void reset_receiver(tbb::flow::reset_flags /*f*/) { my_count = 0; }
-#else
-    /*override*/void reset_receiver() { my_count = 0; }
+    /*override*/void clear_predecessors() { my_count = 0; };
 #endif
-
+    /*override*/void reset_receiver(tbb::flow::reset_flags /*f*/) { my_count = 0; }
 };
 
 //! Counts the number of puts received
@@ -348,7 +349,7 @@ struct harness_mapped_receiver : public tbb::flow::receiver<T>, NoCopy {
       } else {
           ++my_count;
       }
-      return const_cast<tbb::task *>(tbb::flow::interface7::SUCCESSFULLY_ENQUEUED);
+      return const_cast<tbb::task *>(tbb::flow::interface8::SUCCESSFULLY_ENQUEUED);
     }
 
     void validate() {
@@ -363,15 +364,21 @@ struct harness_mapped_receiver : public tbb::flow::receiver<T>, NoCopy {
         }
     }
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+    typedef typename tbb::flow::receiver<T>::built_predecessors_type built_predecessors_type;
+    built_predecessors_type mbp;
+    /*override*/ built_predecessors_type &built_predecessors() { return mbp; }
     typedef typename tbb::flow::receiver<T>::predecessor_list_type predecessor_list_type;
     /*override*/void internal_add_built_predecessor(tbb::flow::sender<T> &) {}
     /*override*/void internal_delete_built_predecessor(tbb::flow::sender<T> &) {}
     /*override*/void copy_predecessors(predecessor_list_type &) { }
     /*override*/size_t predecessor_count() { return 0; }
-    /*override*/void reset_receiver(tbb::flow::reset_flags /*f*/) { my_count = 0; if(my_map) delete my_map; my_map = new map_type; }
-#else
-    /*override*/void reset_receiver() { my_count = 0; if(my_map) delete my_map; my_map = new map_type; }
+    /*override*/void clear_predecessors() { my_count = 0; };
 #endif
+    /*override*/void reset_receiver(tbb::flow::reset_flags /*f*/) {
+        my_count = 0;
+        if(my_map) delete my_map;
+        my_map = new map_type;
+    }
 
 };
 
@@ -410,9 +417,13 @@ struct harness_counting_sender : public tbb::flow::sender<T>, NoCopy {
 
 #if TBB_PREVIEW_FLOW_GRAPH_FEATURES
     typedef typename tbb::flow::sender<T>::successor_list_type successor_list_type;
+    typedef typename tbb::flow::sender<T>::built_successors_type built_successors_type;
+    built_successors_type bst;
+    /*override*/ built_successors_type &built_successors() { return bst; }
     /* override */ void internal_add_built_successor( successor_type &) {}
     /* override */ void internal_delete_built_successor( successor_type &) {}
     /* override */ void copy_successors(successor_list_type &) { }
+    /*override*/ void clear_successors() { my_receiver = NULL; }
     /* override */ size_t successor_count() { return 0; }
 #endif
 
@@ -528,6 +539,7 @@ found_it:
     }
     return true;
 }
+#endif  /* TBB_PREVIEW_FLOW_GRAPH_FEATURES */
 
 template<typename T, typename BufferType>
 void test_resets() {
@@ -632,7 +644,7 @@ void test_resets() {
             serial_fn_state0 = 0;  // release the function_node.
             g.wait_for_all();  // wait for all the tasks to complete.
         }
-        g.reset(tbb::flow::rf_extract);
+        g.reset(tbb::flow::rf_clear_edges);
         // test that no one is a successor to the buffer now.
         serial_fn_state0 = 1;  // let the function_node go if it gets an input message
         b0.try_put((T)23);
@@ -643,6 +655,8 @@ void test_resets() {
     }
 }
 
+#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+
 template< typename NODE_TYPE >
 class test_buffer_base_extract {
 protected:
