@@ -1,35 +1,26 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
 
-    This file is part of Threading Building Blocks.
+    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
+    you can redistribute it and/or modify it under the terms of the GNU General Public License
+    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
+    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
+    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+    See  the GNU General Public License for more details.   You should have received a copy of
+    the  GNU General Public License along with Threading Building Blocks; if not, write to the
+    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
 
-    Threading Building Blocks is free software; you can redistribute it
-    and/or modify it under the terms of the GNU General Public License
-    version 2 as published by the Free Software Foundation.
-
-    Threading Building Blocks is distributed in the hope that it will be
-    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with Threading Building Blocks; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-
-    As a special exception, you may use this file as part of a free software
-    library without restriction.  Specifically, if other files instantiate
-    templates or use macros or inline functions from this file, or you compile
-    this file and link it with other files to produce an executable, this
-    file does not by itself cause the resulting executable to be covered by
-    the GNU General Public License.  This exception does not however
-    invalidate any other reasons why the executable file might be covered by
-    the GNU General Public License.
+    As a special exception,  you may use this file  as part of a free software library without
+    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
+    functions from this file, or you compile this file and link it with other files to produce
+    an executable,  this file does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however invalidate any other
+    reasons why the executable file might be covered by the GNU General Public License.
 */
 
 // undefine __TBB_CPF_BUILD to simulate user's setup
 #undef __TBB_CPF_BUILD
 
-#define TBB_PREVIEW_TASK_ARENA 1
 #define TBB_PREVIEW_LOCAL_OBSERVER 1
 #define __TBB_EXTRA_DEBUG 1
 
@@ -48,6 +39,8 @@
 #include <cstdlib>
 #include <cstdio>
 
+#include "harness_fp.h"
+
 #include "tbb/task_arena.h"
 #include "tbb/task_scheduler_observer.h"
 #include "tbb/task_scheduler_init.h"
@@ -58,7 +51,6 @@
 #include "harness_assert.h"
 #include "harness.h"
 #include "harness_barrier.h"
-#include "harness_concurrency_tracker.h"
 
 #if _MSC_VER
 // plays around __TBB_NO_IMPLICIT_LINKAGE. __TBB_LIB_NAME should be defined (in makefiles)
@@ -120,8 +112,8 @@ void InitializeAndTerminate( int maxthread ) {
 typedef tbb::blocked_range<int> Range;
 
 Harness::SpinBarrier our_barrier;
-
-static tbb::enumerable_thread_specific<int> local_id, old_id, slot_id(-1);
+// slot_id value: -1 is reserved by current_slot(), -2 is set in on_scheduler_exit() below
+static tbb::enumerable_thread_specific<int> local_id, old_id, slot_id(-3);
 void ResetTLS() {
     local_id.clear();
     old_id.clear();
@@ -131,46 +123,41 @@ void ResetTLS() {
 class ConcurrencyTrackingBody {
 public:
     void operator() ( const Range& ) const {
-        ASSERT(slot_id.local() == tbb::task_arena::current_slot(), NULL);
-        Harness::ConcurrencyTracker ct;
-        for ( volatile int i = 0; i < 100000; ++i )
+        ASSERT(slot_id.local() == tbb::task_arena::current_thread_index(), NULL);
+        for ( volatile int i = 0; i < 50000; ++i )
             ;
     }
 };
 
 class ArenaObserver : public tbb::task_scheduler_observer {
     int myId;
-    tbb::atomic<int> myTrappedSlot;
     /*override*/
     void on_scheduler_entry( bool is_worker ) {
-        REMARK("a %s #%p is entering arena %d from %d\n", is_worker?"worker":"master", &local_id.local(), myId, local_id.local());
+        REMARK("a %s #%p is entering arena %d from %d on slot %d\n", is_worker?"worker":"master",
+                &local_id.local(), myId, local_id.local(),
+                tbb::task_arena::current_thread_index());
         ASSERT(!old_id.local(), "double-call to on_scheduler_entry");
         old_id.local() = local_id.local();
         ASSERT(old_id.local() != myId, "double-entry to the same arena");
         local_id.local() = myId;
-        slot_id.local() = tbb::task_arena::current_slot();
-        if(is_worker) ASSERT(tbb::task_arena::current_slot()>0, NULL);
-        else ASSERT(tbb::task_arena::current_slot()==0, NULL);
+        slot_id.local() = tbb::task_arena::current_thread_index();
+        if(is_worker) ASSERT(tbb::task_arena::current_thread_index()>0, NULL);
+        else ASSERT(tbb::task_arena::current_thread_index()==0, NULL);
     }
     /*override*/
     void on_scheduler_exit( bool is_worker ) {
-        REMARK("a %s #%p is leaving arena %d to %d\n", is_worker?"worker":"master", &local_id.local(), myId, old_id.local());
+        REMARK("a %s #%p is leaving arena %d to %d\n", is_worker?"worker":"master",
+                &local_id.local(), myId, old_id.local());
         ASSERT(local_id.local() == myId, "nesting of arenas is broken");
-        ASSERT(slot_id.local() == tbb::task_arena::current_slot(), NULL);
-        slot_id.local() = -1;
+        ASSERT(slot_id.local() == tbb::task_arena::current_thread_index(), NULL);
+        slot_id.local() = -2;
         local_id.local() = old_id.local();
         old_id.local() = 0;
     }
-    /*override*/
-    bool on_scheduler_leaving() {
-        ASSERT(slot_id.local() == tbb::task_arena::current_slot(), NULL);
-        return tbb::task_arena::current_slot() >= myTrappedSlot;
-    }
 public:
-    ArenaObserver(tbb::task_arena &a, int id, int trap = 0) : tbb::task_scheduler_observer(a) {
+    ArenaObserver(tbb::task_arena &a, int id) : tbb::task_scheduler_observer(a) {
         ASSERT(id, NULL);
         myId = id;
-        myTrappedSlot = trap;
         observe(true);
     }
     ~ArenaObserver () {
@@ -185,7 +172,7 @@ struct AsynchronousWork : NoAssign {
     : my_barrier(a_barrier), my_is_blocking(blocking) {}
     void operator()() const {
         ASSERT(local_id.local() != 0, "not in explicit arena");
-        tbb::parallel_for(Range(0,500), ConcurrencyTrackingBody(), tbb::simple_partitioner());
+        tbb::parallel_for(Range(0,500), ConcurrencyTrackingBody(), tbb::simple_partitioner(), *tbb::task::self().group());
         if(my_is_blocking) my_barrier.timed_wait(10); // must be asynchronous to master thread
         else my_barrier.signal_nowait();
     }
@@ -259,6 +246,7 @@ public:
         : my_a(a), my_b(b) {}
     void operator()(int idx) const {
         tbb::empty_task* root_task = new(tbb::task::allocate_root()) tbb::empty_task;
+        my_b.timed_wait(10); // increases chances for task_arena initialization contention
         for( int i=0; i<100; ++i) {
             root_task->set_ref_count(2);
             my_a.enqueue(Runner(root_task));
@@ -270,6 +258,66 @@ public:
     }
 };
 
+class MultipleMastersPart4 : NoAssign {
+    tbb::task_arena &my_a;
+    Harness::SpinBarrier &my_b;
+    tbb::task_group_context *my_ag;
+
+    struct Getter : NoAssign {
+        tbb::task_group_context *& my_g;
+        Getter(tbb::task_group_context *&a_g) : my_g(a_g) {}
+        void operator()() const {
+            my_g = tbb::task::self().group();
+        }
+    };
+    struct Checker : NoAssign {
+        tbb::task_group_context *my_g;
+        Checker(tbb::task_group_context *a_g) : my_g(a_g) {}
+        void operator()() const {
+            ASSERT(my_g == tbb::task::self().group(), NULL);
+            tbb::task *t = new( tbb::task::allocate_root() ) tbb::empty_task;
+            ASSERT(my_g == t->group(), NULL);
+            tbb::task::destroy(*t);
+        }
+    };
+    struct NestedChecker : NoAssign {
+        const MultipleMastersPart4 &my_body;
+        NestedChecker(const MultipleMastersPart4 &b) : my_body(b) {}
+        void operator()() const {
+            tbb::task_group_context *nested_g = tbb::task::self().group();
+            ASSERT(my_body.my_ag != nested_g, NULL);
+            tbb::task *t = new( tbb::task::allocate_root() ) tbb::empty_task;
+            ASSERT(nested_g == t->group(), NULL);
+            tbb::task::destroy(*t);
+            my_body.my_a.enqueue(Checker(my_body.my_ag));
+        }
+    };
+public:
+    MultipleMastersPart4( tbb::task_arena &a, Harness::SpinBarrier &b) : my_a(a), my_b(b) {
+        my_a.execute(Getter(my_ag));
+    }
+    // NativeParallelFor's functor
+    void operator()(int) const {
+        my_a.execute(*this);
+    }
+    // Arena's functor
+    void operator()() const {
+        Checker check(my_ag);
+        check();
+        tbb::task_arena nested(1,1);
+        nested.execute(NestedChecker(*this)); // change arena
+        tbb::parallel_for(tbb::blocked_range<int>(0,1),*this); // change group context only
+        my_b.timed_wait(10);
+        my_a.execute(check);
+        check();
+    }
+    // parallel_for's functor
+    void operator()(const tbb::blocked_range<int> &) const {
+        NestedChecker(*this)();
+        my_a.execute(Checker(my_ag)); // restore arena context
+    }
+};
+
 void TestMultipleMasters(int p) {
     {
         REMARK("multiple masters, part 1\n");
@@ -297,9 +345,130 @@ void TestMultipleMasters(int p) {
         // "Oversubscribe" the arena by 1 master thread
         NativeParallelFor( p+1, MultipleMastersPart3(a, barrier) );
         a.debug_wait_until_empty();
+    } {
+        int c = p%3? (p%2? p : 2) : 3;
+        REMARK("multiple masters, part 4: contexts, arena(%d)\n", c);
+        tbb::task_arena a(c, 1);
+        ArenaObserver o(a, c);
+        Harness::SpinBarrier barrier(c);
+        MultipleMastersPart4 test(a, barrier);
+        NativeParallelFor(p, test);
+        a.debug_wait_until_empty();
     }
 }
 
+#include <sstream>
+#if TBB_USE_EXCEPTIONS
+#include <stdexcept>
+#include "tbb/tbb_exception.h"
+#endif
+
+struct TestArenaEntryBody : FPModeContext {
+    tbb::atomic<int> &my_stage; // each execute increases it
+    std::stringstream my_id;
+    bool is_caught, is_expected;
+    enum { arenaFPMode = 1 };
+
+    TestArenaEntryBody(tbb::atomic<int> &s, int idx, int i)  // init thread-specific instance
+    :   FPModeContext(idx+i)
+    ,   my_stage(s)
+    ,   is_caught(false)
+    ,   is_expected( (idx&(1<<i)) != 0 && (TBB_USE_EXCEPTIONS) != 0 )
+    {
+        my_id << idx << ':' << i << '@';
+    }
+    void operator()() { // inside task_arena::execute()
+        // synchronize with other stages
+        int stage = my_stage++;
+        int slot = tbb::task_arena::current_thread_index();
+        ASSERT(slot >= 0 && slot <= 1, "master or the only worker");
+        // wait until the third stage is delegated and then starts on slot 0
+        while(my_stage < 2+slot) __TBB_Yield();
+        // deduct its entry type and put it into id, it helps to find source of a problem
+        my_id << (stage < 3 ? (tbb::task_arena::current_thread_index()?
+                              "delegated_to_worker" : stage < 2? "direct" : "delegated_to_master")
+                            : stage == 3? "nested_same_ctx" : "nested_alien_ctx");
+        REMARK("running %s\n", my_id.str().c_str());
+        AssertFPMode(arenaFPMode);
+        if(is_expected)
+            __TBB_THROW(std::logic_error(my_id.str()));
+        // no code can be put here since exceptions can be thrown
+    }
+    void on_exception(const char *e) { // outside arena, in catch block
+        is_caught = true;
+        REMARK("caught %s\n", e);
+        ASSERT(my_id.str() == e, NULL);
+        assertFPMode();
+    }
+    void after_execute() { // outside arena and catch block
+        REMARK("completing %s\n", my_id.str().c_str() );
+        ASSERT(is_caught == is_expected, NULL);
+        assertFPMode();
+    }
+};
+
+class ForEachArenaEntryBody : NoAssign {
+    tbb::task_arena &my_a; // expected task_arena(2,1)
+    tbb::atomic<int> &my_stage; // each execute increases it
+    int my_idx;
+
+public:
+    ForEachArenaEntryBody(tbb::task_arena &a, tbb::atomic<int> &c)
+    : my_a(a), my_stage(c), my_idx(0) {}
+
+    void test(int idx) {
+        my_idx = idx;
+        my_stage = 0;
+        NativeParallelFor(3, *this); // test cross-arena calls
+        ASSERT(my_stage == 3, NULL);
+        my_a.execute(*this); // test nested calls
+        ASSERT(my_stage == 5, NULL);
+    }
+
+    // task_arena functor for nested tests
+    void operator()() const {
+        test_arena_entry(3); // in current task group context
+        tbb::parallel_for(4, 5, *this); // in different context
+    }
+
+    // NativeParallelFor & parallel_for functor
+    void operator()(int i) const {
+        test_arena_entry(i);
+    }
+
+private:
+    void test_arena_entry(int i) const {
+        TestArenaEntryBody scoped_functor(my_stage, my_idx, i);
+        __TBB_TRY {
+            my_a.execute(scoped_functor);
+        }
+#if TBB_USE_EXCEPTIONS
+        catch(tbb::captured_exception &e) {
+            scoped_functor.on_exception(e.what());
+            ASSERT_WARNING(TBB_USE_CAPTURED_EXCEPTION, "Caught captured_exception while expecting exact one");
+        } catch(std::logic_error &e) {
+            scoped_functor.on_exception(e.what());
+            ASSERT(!TBB_USE_CAPTURED_EXCEPTION, "Caught exception of wrong type");
+        } catch(...) { ASSERT(false, "Unexpected exception type"); }
+#endif //TBB_USE_EXCEPTIONS
+        scoped_functor.after_execute();
+    }
+};
+
+void TestArenaEntryConsistency() {
+    REMARK("test arena entry consistency\n" );
+
+    tbb::task_arena a(2,1);
+    tbb::atomic<int> c;
+    ForEachArenaEntryBody body(a, c);
+
+    FPModeContext fp_scope(TestArenaEntryBody::arenaFPMode);
+    a.initialize(); // capture FP settings to arena
+    fp_scope.setNextFPMode();
+
+    for(int i = 0; i < 100; i++) // not less than 32 = 2^5 of entry types
+        body.test(i);
+}
 
 int TestMain () {
     // TODO: a workaround for temporary p-1 issue in market
@@ -312,5 +481,6 @@ int TestMain () {
         TestMultipleMasters( p );
         ResetTLS();
     }
+    TestArenaEntryConsistency();
     return Harness::Done;
 }
